#ifndef __CONFIG_H
#define __CONFIG_H

#define MAX_PATH		256
#define MAX_EPG_POSTMINUTES	30
#define CONFIG_FNAME		"pvrplus.cfg"

enum
{
   MENULANG_BG = 0x06,
   MENULANG_HR, MENULANG_CZ, MENULANG_DK, MENULANG_NL, MENULANG_FI, MENULANG_FR, MENULANG_DE, MENULANG_GR,
   MENULANG_HU, MENULANG_IT, MENULANG_NO, MENULANG_EN, MENULANG_PL, MENULANG_PT, MENULANG_RO, MENULANG_RU,
   MENULANG_RS, MENULANG_SI, MENULANG_ES, MENULANG_SE, MENULANG_TR,
   MENULANG_EE = 0x24, MENULANG_LV, MENULANG_LT
};

enum
{
   MODEL_UNKNOWN=0, MODEL_CHL7DEUC
};

enum
{
   EPG_TRANS_NONE = 0,
   EPG_TRANS_ASCII,
   EPG_TRANS_UTF8,
   MAX_EPG_TRANS
};

enum
{
   TIMESOURCE_NTP = 0,
   TIMESOURCE_EPG,
   MAX_TIMESOURCES
};

enum
{  FREEZE_OFF = 0,
   FREEZE_ON,
   FREEZE_AUTO
};

typedef struct			// main configuration structure
{
  int menu_language;		// OSD menu language from TV
  char path[MAX_PATH];		// path to main module (with ending slash)
  target target1;		// basic recording target from targets[] table
  target target2;		// alternative recording target from targets[] table
  int current_target;		// number of current target (0==1==first, 2=second)
  int useintmem;		// allow to use internal memory for recordings (0=no, 1=yes)
  int source;			// SOURCE key function (0=disabled, 1=HDMI1, 2=HDMI2,...)
  int favch;			// FAV.CH key function (0=disabled, 1=aspect ratio,...)
  int mute;			// MUTE key function (0=mute only, 1=mute/panel off)
  int record_key_lock;		// enable recording lock function ?
  int EPG_transcode;		// translate national characters in EPG ? (EPG_TRANS_NONE, EPG_TRANS_ASCII, EPG_TRANS_UTF8)
  int EPG_postminutes;		// time added to EPG autostop signal [minutes]
  int timezone;			// timezone value based on timezones[] table
  int timesource;		// get time from: 0=NTP/GMT system time, 1=EPG
  int OSD;			// OSD opacity ? (0=disable, 10=10%, 20=20%,... 100=full)
  int OSD_effects;		// OSD effects (0=disable, 1=enable)
  unsigned char alpha;		// OSD transparency [%] (0=off...100=opaque)
  int testwrite;		// test for writing before recording
  int infopane;			// show infopane (0=manual, 1=auto)
  int del_mdb;			// delete mdb file after recording ?
  int del_empty_ts;		// delete empty ts file after recording ?
  unsigned io_buffer;		// internal I/O buffer size (0=default)
  int divide;			// divide recording (0=off, 1=auto, other values - see divides[] table)
  int summary;			// display recording summary (0=off, 1=on)
  int warnings;			// display warnings (0=off, 1=on)
  int recwarnings;		// display periodical warnings that recording is in progress and panel is off (0=off, 1=on)
  int teletext;			// allow enter teletext during recording (0=no, 1=yes)
  int infopanefont;		// infopane font (0=default/DOT18, 1=LCD18, 2=ARIAL13)
  int pre_scripts;		// run pre-recording scripts (0=no, 1=yes)
  int createintdirs;		// create internal directories before recording (0=no, 1=yes)
  int poweroff;			// power off at autostop (0=no, 1=yes)
  int freeze;			// pause (video freeze) function (0=off, 1=on, 2=auto)
  int model;			// TV model type (0=unknown(CIP?), 1=CHL7DEUC)
  int DVRnr;			// number of DVR subdirectory used (found)
  unsigned *KeyInputBase_SendKey;
  unsigned *pKeyInputBase;
} cfgstruct;

#define DEFAULT_CONFIG	{ \
MENULANG_EN,			/* menu_language */	\
"",				/* path */		\
AUTO_TARGET_DEF,		/* target1 */		\
AUTO_TARGET_DEF,		/* target2 */		\
1,				/* current target */	\
0,				/* useintmem */		\
1,				/* source */		\
1,				/* favch */		\
1,				/* mute */		\
0,				/* record_key_lock */	\
EPG_TRANS_ASCII,		/* EPG_transcode */	\
0,				/* EPG_postminutes */	\
1,				/* timezone */		\
TIMESOURCE_NTP,			/* timesource */	\
100,				/* OSD */		\
1,				/* OSD effects */	\
80,				/* alpha */		\
1,				/* testwrite */		\
0,				/* infopane */		\
1,				/* del_mdb */		\
1,				/* del_empty_ts */	\
1732608,			/* io_buffer */		\
1,				/* divide */		\
0,				/* summary */		\
1,				/* warnings */		\
1,				/* recwarnings */	\
1,				/* teletext */		\
0,				/* infopanefont */	\
1,				/* pre_scripts */	\
0,				/* createintdirs */	\
0,				/* poweroff */		\
FREEZE_AUTO,			/* freeze */		\
1,				/* model */		\
0,				/* DVRnr */		\
NULL,				/* KeyInputBase_SendKey */ \
NULL,				/* pKeyInputBase */	\
}

#endif

