// ===========================================================================
// PVR+ 1.3 loader
// ===========================================================================

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <dlfcn.h>
#include <errno.h>
#include <unistd.h>
#include <sys/mman.h>
#include <string.h>
#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include "pvrplus.h"
#include "config.h"
#include "externs.h"
#include "messages.h"

#define MAIN_MODULE			"pvrplus.so"
#define DEBUG_LOG 			"/mtd_ram/PVRplus_debug.log"
#define HELPER_MODULE			"osd_ttf.so"
#define OSD_TTF_Init			(*ex_OSD_TTF_Init)
#define OSD_TTF_WasInit			(*ex_OSD_TTF_WasInit)
#define OSD_TTF_Quit			(*ex_OSD_TTF_Quit)
#define OSD_TTF_OpenFont		(*ex_OSD_TTF_OpenFont)
#define OSD_TTF_CloseFont		(*ex_OSD_TTF_CloseFont)
#define OSD_TTF_RenderUTF8_Blended	(*ex_OSD_TTF_RenderUTF8_Blended)
#define OSD_TTF_SizeUTF8		(*ex_OSD_TTF_SizeUTF8)
#define CR				13
#define LF				10
#define MAX_MESS			103
#define LOCAL_FONTS_SUBDIR		"fonts"	
#define FONTS_SUBDIR			"/mtd_appdata/Font"
#define SCREEN_W			(1920/2)
#define SCREEN_H			(1080/2)
#define SCREEN_BPP			32
#define SCREEN_FLAGS			SDL_HWSURFACE
#define WARNING_DELAY			3000
#define MIN_MESS_ALLOC			6
#define MAX_MESS_ALLOC			(3*1024*1024)

enum
{
   TVMODEL_UNKNOWN = 0,
   TVMODEL_CHL7DEUC
};

int CFG_writable;
const char *tagfile = "/mtd_ram/.pvrplus";
static cfgstruct config = DEFAULT_CONFIG;
int OSD_TTF_on;
unsigned *OSD_handle = NULL;
int last_menulang = -1;
char *mess[MAX_MESS];
char *mess_buffer = NULL;
char *FONT_SHADOW = "shadow.ttf";
const char *OSD_SUBDIR	 = "osd";
const char *ICON_WARNBOX = "warnbox";
static SDL_Surface *screen;
void *ttf_font;
char *CHL7DEUC_TAG	= "CHL7DEUC";

extern int _ZN4TCTv14PreferLanguageEv(int *);

int (*ex_OSD_TTF_Init)(void);
int (*ex_OSD_TTF_WasInit)(void);
void (*ex_OSD_TTF_Quit)(void);
void *(*ex_OSD_TTF_OpenFont)(const char *, int);
void (*ex_OSD_TTF_CloseFont)(void *);
SDL_Surface *(*ex_OSD_TTF_RenderUTF8_Blended)(void *, const char *, SDL_Color);
int (*ex_OSD_TTF_SizeUTF8)(void *, const char *, int *, int *);


void debug(char *text)
{
   FILE *f;
   char s[256];

   sprintf(s, "%s\n", text);
   f = fopen(DEBUG_LOG, "at");
   if (f) { fputs(s, f); fflush(f); fclose(f); }
}

int no_function(unsigned *handle, char *function)
{
   char s[128];
   sprintf(s, "Function [%s] not found:", function);
   debug(s);
   debug(dlerror());
   if (handle) dlclose(handle);
   return 0;
}

int OSD_init_ttf_helper()
{
   char s[MAX_PATH];

   if (OSD_TTF_on) return 1;
   sprintf(s, "%s%s", config.path, HELPER_MODULE);
   if ( (OSD_handle = dlopen(s, RTLD_LAZY)) == NULL ) return 0;
   ex_OSD_TTF_Init = dlsym(OSD_handle, "OSD_TTF_Init");
   if (ex_OSD_TTF_Init == NULL) return no_function(OSD_handle, "OSD_TTF_Init");
   ex_OSD_TTF_WasInit = dlsym(OSD_handle, "OSD_TTF_WasInit");
   if (ex_OSD_TTF_WasInit == NULL) return no_function(OSD_handle, "OSD_TTF_WasInit");
   ex_OSD_TTF_Quit = dlsym(OSD_handle, "OSD_TTF_Quit");
   if (ex_OSD_TTF_Quit == NULL) return no_function(OSD_handle, "OSD_TTF_Quit");
   ex_OSD_TTF_OpenFont = dlsym(OSD_handle, "OSD_TTF_OpenFont");
   if (ex_OSD_TTF_OpenFont == NULL) return no_function(OSD_handle, "OSD_TTF_OpenFont");
   ex_OSD_TTF_CloseFont = dlsym(OSD_handle, "OSD_TTF_CloseFont");
   if (ex_OSD_TTF_CloseFont == NULL) return no_function(OSD_handle, "OSD_TTF_CloseFont");
   ex_OSD_TTF_RenderUTF8_Blended = dlsym(OSD_handle, "OSD_TTF_RenderUTF8_Blended");
   if (ex_OSD_TTF_RenderUTF8_Blended == NULL) return no_function(OSD_handle, "OSD_TTF_RenderUTF8_Blended");
   ex_OSD_TTF_SizeUTF8 = dlsym(OSD_handle, "OSD_TTF_SizeUTF8");
   if (ex_OSD_TTF_SizeUTF8 == NULL) return no_function(OSD_handle, "OSD_TTF_SizeUTF8");
   OSD_TTF_on = 1;
   return 1;
}

int get_menu_language()
{
   int ret;
   ret = _ZN4TCTv14PreferLanguageEv((int *)g_AppTV);
   return ret;
}

char *lang_suffix(int menu_language)
{
   switch (menu_language)
   {
      case MENULANG_PL: return "PL";
      case MENULANG_DE: return "DE";
      case MENULANG_FR: return "FR";
      case MENULANG_BG: return "BG";
      case MENULANG_HR: return "HR";
      case MENULANG_CZ: return "CZ";
      case MENULANG_DK: return "DK";
      case MENULANG_NL: return "NL";
      case MENULANG_FI: return "FI";
      case MENULANG_GR: return "GR";
      case MENULANG_HU: return "HU";
      case MENULANG_IT: return "IT";
      case MENULANG_NO: return "NO";
      case MENULANG_PT: return "PT";
      case MENULANG_RO: return "RO";
      case MENULANG_RU: return "RU";
      case MENULANG_RS: return "RS";
      case MENULANG_SI: return "SI";
      case MENULANG_ES: return "ES";
      case MENULANG_SE: return "SE";
      case MENULANG_TR: return "TR";
      case MENULANG_EE: return "EE";
      case MENULANG_LV: return "LV";
      case MENULANG_LT: return "LT";

      default:		return "EN";
   }
}

void load_messages()
{
   int i, ln;
   char *pos, *stor;
   FILE *f;
   char s[256];
   char mess_fname[MAX_PATH];

   sprintf(mess_fname, "%smessages_%s.txt", config.path, lang_suffix(config.menu_language));
   if (access(mess_fname, F_OK) != 0) sprintf(mess_fname, "%smessages_%s.txt", config.path, lang_suffix(MENULANG_EN));
   struct stat fst;
   unsigned long mess_mem;
   if (stat(mess_fname, &fst) == 0) mess_mem = (unsigned long)fst.st_size;
   else mess_mem = 0L;
   for (i=0; i < MAX_MESS; i++) mess[i] = "?";
   if ((mess_mem < MIN_MESS_ALLOC) || (mess_mem > MAX_MESS_ALLOC)) return;
   mess_buffer = malloc(mess_mem - 2);
   if (mess_buffer)
   {
      stor = mess_buffer;

      f = fopen(mess_fname, "rt");
      if (f)
      {
	 while (!feof(f))
	 {
	    if (fgets(s, sizeof(s)-1, f) == s)
	    {
	       ln = strlen(s);
	       if (ln > 0) for (i=0; i < ln; i++) if ((s[i]==CR) || (s[i]==LF)) s[i]=0;
	       ln = strlen(s);
	       if (ln > 0)
	          if (s[0] != ';')
		  {
		     pos = strstr(s, "=");
		     if (pos > s)
		     {
			*pos = 0;
			i = atoi(s);
			if ((i >= 0) && (i < MAX_MESS))
			{
			   strcpy(stor, pos+1);
			   mess[i] = stor;
			   stor += strlen(stor);
			   *(stor++) = 0;
			}
		     }
		  }
	    }
	 }
	 fclose(f);
      }
   }

   last_menulang = config.menu_language;
}

void unload_messages()
{
   if (mess_buffer) { free(mess_buffer); mess_buffer = NULL; }
}

void loader_show_text(int xpos, int ypos, char *text)
{
   int ln = strlen(text);

   if ((xpos<0) || (ypos<0) || (ln==0)) return;

   if (screen)
   {
      SDL_Surface *ttf_text;
      if (ttf_font)
      {
	 SDL_Rect ttf_rect = {xpos, ypos, SCREEN_W-xpos, SCREEN_H-ypos};
	 ttf_text = OSD_TTF_RenderUTF8_Blended(ttf_font, text, (SDL_Color){200,200,200,0});
	 SDL_BlitSurface(ttf_text, NULL, screen, &ttf_rect);
	 SDL_FreeSurface(ttf_text);
      }
   }
}

void show_loader_WARNING(int msg)
{
   SDL_Rect dstrect;
   SDL_Surface *icon;
   char s[MAX_PATH];

   SDL_Init(SDL_INIT_VIDEO);

   OSD_init_ttf_helper();

   if (!OSD_TTF_WasInit()) if (OSD_TTF_Init() != 0)
   {
      SDL_Quit();
      if (OSD_handle) dlclose(OSD_handle);
      return;
   }

   if (ttf_font == NULL)
   {
      sprintf(s, "%s/%s", FONTS_SUBDIR, FONT_SHADOW);
      if (access(s, F_OK) != 0) sprintf(s, "%s%s/%s", config.path, LOCAL_FONTS_SUBDIR, FONT_SHADOW);
      ttf_font = OSD_TTF_OpenFont(s, 18);
   }

   screen = SDL_SetVideoMode(SCREEN_W, SCREEN_H, SCREEN_BPP, SCREEN_FLAGS);
   if (screen)
   {
      sprintf(s, "%s%s/%s.png", config.path, OSD_SUBDIR, ICON_WARNBOX);
      icon = IMG_Load(s);
      if (icon)
      {
	 dstrect.x = (SCREEN_W - icon->w) >> 1;
	 dstrect.y = (SCREEN_H - icon->h) >> 1;
	 dstrect.w = icon->w;
	 dstrect.h = icon->h;
	 SDL_BlitSurface(icon, NULL, screen, &dstrect);
	 SDL_FreeSurface(icon); icon = NULL;
      }

      config.menu_language = get_menu_language();
      if (config.menu_language != last_menulang) load_messages();
      loader_show_text(dstrect.x + 90, dstrect.y + 40, mess[msg]);
      SDL_Flip(screen);
      SDL_Delay(WARNING_DELAY);
   }

   if (ttf_font) { OSD_TTF_CloseFont(ttf_font); ttf_font = NULL; }
   SDL_Quit();
   if (OSD_handle) dlclose(OSD_handle);
}

static void cacheflush(void *start, size_t size)
{
   asm(
       "mov r0, %0\n"
       "mov r1, %1\n"
       "mov r2, #0\n"
       "mov r7, #0xf0000\n"
       "add r7, #0x02\n"
       "swi #0\n" ::
       "g"((long)start), "g"((long)start+size) :
       "r0", "r1", "r2", "r7");
}

int get_model()
{
   FILE *f;
   char s[50];
   int b;
   int ret = TVMODEL_UNKNOWN;

   f = fopen("/mtd_rwarea/Version.1", "rb");
   if (f)
   {
       b = fread(&s, 1, sizeof(s)-2, f);
       if (b > sizeof(CHL7DEUC_TAG))
       {
	  if (b < sizeof(s)) s[b] = 0; else s[sizeof(s)-1] = 0;
          if (strstr(s, CHL7DEUC_TAG)) ret = TVMODEL_CHL7DEUC;
       }
       fclose(f);
       if (ret) return ret;
   }

   f = fopen("/mtd_rwarea/Version.0", "rb");
   if (f)
   {
       b = fread(&s, 1, sizeof(s)-2, f);
       if (b > sizeof(CHL7DEUC_TAG))
       {
	  if (b < sizeof(s)) s[b] = 0; else s[sizeof(s)-1] = 0;
          if (strstr(s, CHL7DEUC_TAG)) ret = TVMODEL_CHL7DEUC;
       }
       fclose(f);
       if (ret) return ret;
   }

   return TVMODEL_UNKNOWN;
}

int injector(const char *path)
{
	char s[MAX_PATH];

	debug("injector() started");
	unsigned *handle=NULL;
	if ( (handle = dlopen(NULL, RTLD_LAZY | RTLD_GLOBAL)) == NULL )
	{
	   debug("cannot open self:");
	   debug(dlerror());
	   return 0;
	}

	debug("self open");
	unsigned *spI_Stream_Record = dlsym(handle, "spI_Stream_Record");
	if (spI_Stream_Record == NULL) return no_function(handle, "spI_Stream_Record");
	unsigned *KeyCommon_SendKeyPressInput = dlsym(handle, "_ZN9KeyCommon17SendKeyPressInputEii");
	if (KeyCommon_SendKeyPressInput == NULL ) return no_function(handle, "_ZN9KeyCommon17SendKeyPressInputEii");
	unsigned *KeyCommon_SendKeyReleaseInput = dlsym(handle, "_ZN9KeyCommon19SendKeyReleaseInputEi");
	if (KeyCommon_SendKeyReleaseInput == NULL ) return no_function(handle, "_ZN9KeyCommon19SendKeyReleaseInputEi");
	config.KeyInputBase_SendKey = dlsym(handle,"_ZN14SsKeyInputBase7SendKeyEi");
	if (config.KeyInputBase_SendKey == NULL ) return no_function(handle, "_ZN14SsKeyInputBase7SendKeyEi");
	config.pKeyInputBase = dlsym(handle,"_ZN8SsObject14m_poObjectListE");
	if (config.pKeyInputBase == NULL ) return no_function(handle, "_ZN8SsObject14m_poObjectListE");
	void *g_AppTV = dlsym(handle, "g_AppTV");
	if (g_AppTV == NULL) return no_function(handle, "g_AppTV");
	dlclose(handle);
	sprintf(s, "%s%s", path, MAIN_MODULE);
	debug("loading main module:");
	debug(s);

	if ( (handle = dlopen(s, RTLD_NOW | RTLD_LOCAL | RTLD_NODELETE)) == NULL )
	{
	   debug("cannot open main module:");
	   debug(dlerror());
	   return 0;
	}

	unsigned *ex_injection_press = dlsym(handle, "injection_press");
	if (ex_injection_press == NULL) return no_function(handle, "injection_press");
	unsigned *ex_injection_release = dlsym(handle, "injection_release");
	if (ex_injection_release == NULL) return no_function(handle, "injection_release");
	void (*ex_init_config)(cfgstruct *);
	*(void **)(&ex_init_config) = dlsym(handle, "init_config");
	if (ex_init_config == NULL ) return no_function(handle, "init_config");
	dlclose(handle);
	debug("main module handle closed");
	(*ex_init_config)(&config);
	debug("config copied");
	int page = ((unsigned) spI_Stream_Record) & ~0x0FFF;
	mprotect((void *) page, 8192, PROT_READ | PROT_WRITE | PROT_EXEC);
	spI_Stream_Record[0x11] = 0xE3A00001;
	spI_Stream_Record[0x17] = 0xE3A00002;
	spI_Stream_Record[0x1B] = 0xE3A00003;
	mprotect((void *) page, 8192, PROT_READ | PROT_EXEC);
 	page = ((unsigned) KeyCommon_SendKeyPressInput) & ~0x0FFF;
	mprotect((void *) page, 8192, PROT_READ | PROT_WRITE | PROT_EXEC);

	if (KeyCommon_SendKeyPressInput[4] == 0xE51FF004)
	{
	   debug("old injection_press detected");
	   int page2 = ((unsigned) ex_injection_press) & ~0x0FFF;
	   mprotect((void *) page2, 8192, PROT_READ | PROT_WRITE | PROT_EXEC);
	   ex_injection_press[9]  = KeyCommon_SendKeyPressInput[4];
	   ex_injection_press[10] = KeyCommon_SendKeyPressInput[5];
	   mprotect((void *) page2, 8192, PROT_READ | PROT_EXEC);
	   cacheflush(ex_injection_press+9, 8);
	}
	else debug ("first injection_press");

	if (KeyCommon_SendKeyReleaseInput[4] == 0xE51FF004)
	{
	   debug("old injection_release detected");
	   int page2 = ((unsigned) ex_injection_release) & ~0x0FFF;
	   mprotect((void *) page2, 8192, PROT_READ | PROT_WRITE | PROT_EXEC);
	   ex_injection_release[7]  = KeyCommon_SendKeyReleaseInput[4];
	   ex_injection_release[8] = KeyCommon_SendKeyReleaseInput[5];
	   mprotect((void *) page2, 8192, PROT_READ | PROT_EXEC);
	   cacheflush(ex_injection_release+7, 8);
	}
	else debug ("first injection_release");

	KeyCommon_SendKeyPressInput[4] = 0xE51FF004;
	KeyCommon_SendKeyPressInput[5] = (unsigned)ex_injection_press;
	KeyCommon_SendKeyReleaseInput[4] = 0xE51FF004;
	KeyCommon_SendKeyReleaseInput[5] = (unsigned)ex_injection_release;
	mprotect((void *) page, 8192, PROT_READ | PROT_EXEC);
	cacheflush(spI_Stream_Record+0x11, 4);
	cacheflush(spI_Stream_Record+0x17, 4);
	cacheflush(spI_Stream_Record+0x1B, 4);
	cacheflush(KeyCommon_SendKeyPressInput+4, 8);
	cacheflush(KeyCommon_SendKeyReleaseInput+4, 8);
	debug("Injection complete");
	return 1;
}

void load_config(const char *path)
{
   FILE *cfg;
   int ln;
   char tmp_path[MAX_PATH];
   char s[MAX_PATH];

   CFG_writable = 1;
   strcpy(tmp_path, path);
   ln = strlen(tmp_path);
   if (ln == 0) strcpy(tmp_path, "./");
   else if (tmp_path[ln-1] != '/') strcat(tmp_path, "/");

   sprintf(s, "%s%s", tmp_path, CONFIG_FNAME);
   cfg = fopen(s, "rb");
   if (cfg)
   {
      ln = fread(&config, sizeof(config), 1, cfg);
      fclose(cfg);

      if (access(s, W_OK) != 0) CFG_writable = 0;
      else
      {
	 cfg = fopen(s, "ab");
	 if (cfg) fclose(cfg); else CFG_writable = 0;
      }
   }

   strcpy(config.path, tmp_path);
   config.menu_language = get_menu_language();
   config.model = get_model();
}

int already_loaded(char *main_module)
{
   FILE *f;
   char cmd[256];
   const char *logfile = "/mtd_ram/loader_test.log";

   if (access(tagfile, F_OK) == 0) return 1;
   remove(logfile);
   sprintf(cmd, "cat /proc/`pidof exeDSP`/maps | grep %s >%s", main_module, logfile);

   if (system(cmd) != -1)
   {
      sleep(1);
      f = fopen(logfile, "rt");
      if (f)
      {
	 if (fgets(cmd, sizeof(cmd), f))
	 {
	    if (strstr(cmd, main_module))
	    {
	       fclose(f); remove(logfile);
	       return 1;
	    }
	 }
	 fclose(f);
      }
      remove(logfile);
   }
   return 0;
}

int Game_Main(const char *path, const char *udn __attribute__ ((unused)))
{
	load_config(path);

	if (!already_loaded(MAIN_MODULE))
	{
	   if (injector(path))
	   {
	      remove(DEBUG_LOG);
	      if (!CFG_writable) show_loader_WARNING(MSG_CFG_NOTWRITABLE);

	      FILE *f;
	      f = fopen(tagfile, "wt"); if (f) fclose(f);
	   }
	   else show_loader_WARNING(MSG_ERR_INJECT);

	   unload_messages();
	   return 1;
	}
	else
	{
	   debug("PVRplus already installed");
	   show_loader_WARNING(MSG_ERR_ALREADY);
	}

	unload_messages();
	return 1;
}

