// ===========================================================================
// PVR+ 1.3 main module
// ===========================================================================

#include <sys/vfs.h>
#include <sys/statvfs.h>
#include <sys/stat.h>
#include <dlfcn.h>
#include <errno.h>
#include <unistd.h>
#include <pthread.h>
#include <signal.h>
#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include "buttons.h"
#include "esaving.h"
#include "pvrplus.h"
#include "config.h"
#include "messages.h"
#include "osd.h"
#include "externs.h"

#define MAX_REC_HISTORY			10
#define CR				13
#define LF				10
#define MAX_MESS			103
#define DEBUG_LOG 			"/mtd_ram/PVRplus_debug.log"
#define MAX_ALL_TARGETS			20
#define MAX_LINE			255
#define MAX_TIMEZONES			3
#define REPLACEMENT			'_'
#define MAX_PROGNAME			100
#define NAMEBUFSIZE			256
#define EPG_year			EPG_datetime[0]
#define EPG_month			EPG_datetime[1]
#define EPG_day				EPG_datetime[2]
#define EPG_hour			EPG_datetime[3]
#define EPG_minute			EPG_datetime[4]
#define HELPER_MODULE			"osd_ttf.so"
#define OSD_TTF_Init			(*ex_OSD_TTF_Init)
#define OSD_TTF_WasInit			(*ex_OSD_TTF_WasInit)
#define OSD_TTF_Quit			(*ex_OSD_TTF_Quit)
#define OSD_TTF_OpenFont		(*ex_OSD_TTF_OpenFont)
#define OSD_TTF_CloseFont		(*ex_OSD_TTF_CloseFont)
#define OSD_TTF_RenderUTF8_Blended	(*ex_OSD_TTF_RenderUTF8_Blended)
#define OSD_TTF_SizeUTF8		(*ex_OSD_TTF_SizeUTF8)
#define LOCAL_FONTS_SUBDIR		"fonts"
#define FONTS_SUBDIR			"/mtd_appdata/Font"
#define FI_COLS				16
#define FI_BOLD32_HEIGHT		32
#define FI_SHADOW18_HEIGHT		20
#define FI_DOT18_HEIGHT			18
#define FI_BOLD32_MAX_WIDTH		32
#define FI_SHADOW20_MAX_WIDTH		18
#define FI_DOT18_MAX_WIDTH		12
#define FI_LCD18_MAX_WIDTH		12
#define FI_BOLD_BPP			4
#define FI_DOT_BPP			1
#define fi_lcd18 			fi_dot18
#define MAX_INFOPANEFONT		3
#define INFOPANE_TOP_LINE		390
#define RED_MINUTES_FREE		4
#define TOO_BIG_BITRATE			8000
#define FREE_LEVELS			84
#define WARNING_LINE			505
#define INFOPANE_LINE			(FI_DOT18_HEIGHT + 10)
#define DIM_SPEED			8
#define WARNINFO_X			140
#define WARNINFO_Y			12
#define WARNINFO_BAR_ALPHA		128
#define WARNINFO_BAR_COLOR		8
#define MAX_SETUP_ITEMS			9	
#define SETUP_ITEMS			(SI_END - SI_START)
#define SCRIPT_MAX_TIME			10000
#define SCRIPT_TIME_STEP		100
#define DP_INST0			0
#define DP_INST1			1
#define PAUSE_ON			1
#define PAUSE_OFF			0
#define REC_TIMER_RATE			200
#define MIN_MESS_ALLOC			6
#define MAX_MESS_ALLOC			(3*1024*1024)
#define free_blocks_minimum		1L

enum
{
   SCRIPT_FAILED = -1,
   SCRIPT_NOT_FOUND = 0,
   SCRIPT_SUCCESS
};
enum
{
   SHADOW18=18, SHADOW20=20, SHADOW26=26, BOLD32=32, DOT18=118, LCD18=218
};
enum
{
   Apl=1,  Cpl, Epl, Lpl, Npl, Opl, Spl, Xpl, Zpl,  Ade, Ode, Ude,  eGfr, eAfr, eCfr,
   SPC=16,
   apl=17, cpl, epl, lpl, npl, opl, spl, xpl, zpl,  ade, ode, ude, ssde
};

struct
{
  char *name;
  int font;
  int max_width;
} infopanefonts[MAX_INFOPANEFONT] =
{
   {"Standard",		SHADOW20,	FI_SHADOW20_MAX_WIDTH},
   {"Dot Matrix",	DOT18,		FI_DOT18_MAX_WIDTH},
   {"LCD",		LCD18,		FI_LCD18_MAX_WIDTH}
};
struct
{
   char *name;
   int winter_time;
   int summer_time;
} timezones[MAX_TIMEZONES] =
{
   {"GMT (UTC)", 0,	 0},
   {"CET/CEST",	 1,	 2},
   {"EET/EEST",	 2,	 3},
};
struct
{
   char ts_filename[MAX_PATH];
   char mdb_filename[MAX_PATH];
} rec_hist_files[MAX_REC_HISTORY];
struct statfs fs;
const static struct _U2A { unsigned short unicode; char *ascii; } unicode2ascii[] = {
// Polish
{0x0501, "a"},{0x0401, "A"},{0x0701, "c"},{0x0601, "C"},{0x1901, "e"},{0x1801, "E"},{0x4201, "l"},{0x4101, "L"},
{0x4401, "n"},{0x4301, "N"},{0xF300, "o"},{0xD300, "O"},{0x5B01, "s"},{0x5A01, "S"},{0x7A01, "z"},{0x7901, "Z"},
{0x7C01, "z"},{0x7B01, "Z"},
// Hungarian (2 variants)
{0x5001, "OE"},{0x5101, "oe"},{0xD500, "OE"},{0xF500, "oe"},{0x7001, "UE"},{0x7101, "ue"},{0xDB00, "UE"},{0xFB00, "ue"},
// German + Swedish
{0xE400, "ae"},{0xC400, "AE"},{0xF600, "oe"},{0xD600, "OE"},{0xFC00, "ue"},{0xDC00, "UE"},{0xDF00, "ss"},
// Swedish + Danish
{0xE500, "a"},{0xC500, "A"},
// Danish
{0xE600, "ae"},{0xC600, "AE"},{0xF800, "0"},{0xD800, "0"},
// Russian
{0x1004, "A"},{0x3004, "a"},{0x1104, "B"},{0x3104, "b"},{0x1204, "V"},{0x3204, "v"},{0x1304, "G"},{0x3304, "g"},
{0x1404, "D"},{0x3404, "d"},{0x1504, "E"},{0x3504, "e"},{0x0104, "E"},{0x5104, "e"},{0x1604, "ZH"},{0x3604, "zh"},
{0x1704, "Z"},{0x3704, "z"},{0x1804, "I"},{0x3804, "i"},{0x1904, "I"},{0x3904, "i"},{0x1A04, "K"},{0x3A04, "k"},
{0x1B04, "L"},{0x3B04, "l"},{0x1C04, "M"},{0x3C04, "m"},{0x1D04, "N"},{0x3D04, "n"},{0x1E04, "O"},{0x3E04, "o"},
{0x1F04, "P"},{0x3F04, "p"},{0x2004, "R"},{0x4004, "r"},{0x2104, "S"},{0x4104, "s"},{0x2204, "T"},{0x4204, "t"},
{0x2304, "U"},{0x4304, "u"},{0x2404, "F"},{0x4404, "f"},{0x2504, "KH"},{0x4504, "kh"},{0x2604, "TS"},{0x4604, "ts"},
{0x2704, "CH"},{0x4704, "ch"},{0x2804, "SH"},{0x4804, "sh"},{0x2904, "SHCH"},{0x4904, "shch"},{0x2A04, ""},{0x4A04, ""},
{0x2B04, "Y"},{0x4B04, "y"},{0x2C04, ""},{0x4C04, ""},{0x2D04, "E"},{0x4D04, "e"},{0x2E04, "YU"},{0x4E04, "yu"},
{0x2F04, "YA"},{0x4F04, "ya"},
// ??? (ìéÌÉ)
{0x1B01, "e"},{0x1A01, "E"},{0x1501, "e"},{0x1401, "E"},{0xE900, "e"},{0xC900, "E"},
// ??? (áÁ)
{0xE100, "a"},{0xC100, "A"},
// ??? (úùÚÙ)
{0x6F01, "u"},{0x6E01, "U"},{0xFA00, "u"},{0xDA00, "U"},
// ??? (ïÏòÒ)
{0x6501, "t"},{0x6401, "T"},{0x0F01, "d"},{0x0E01, "D"},{0x4801, "n"},{0x4701, "N"},
// ??? (èÈøØýÝíÍ)
{0x9A00, "s"},{0x8A00, "S"},{0x6101, "s"},{0x6001, "S"},{0x0D01, "c"},{0x0C01, "C"},{0x5901, "r"},{0x5801, "R"},
{0x9E00, "z"},{0x8E00, "Z"},{0x7E01, "z"},{0x7D01, "Z"},{0xFD00, "y"},{0xDD00, "Y"},{0xED00, "i"},{0xCD00, "I"},
{0x0000, ""} // end of table
};
static unsigned utf8_table1[] = {0x7f, 0x7ff, 0xffff, 0x1fffff, 0x3ffffff, 0x7fffffff};
static unsigned char utf8_table2[] = {0, 0xc0, 0xe0, 0xf0, 0xf8, 0xfc};
#define UTF8_TABLE1_SIZE (sizeof(utf8_table1) / sizeof(utf8_table1[0]))
static cfgstruct config = DEFAULT_CONFIG;
const static char *ts_file_pattern="%s/R%s%03d.ts";
const static char *mdb_file_pattern="%s/R%s%03d.mdb";
int no_dir1;
int is_recording;
int recdispmode;
SDL_TimerID rec_timer_id;
time_t rec_starttime;
time_t rec_partstarttime;
int part_number;
unsigned bitrate;
unsigned long free_minutes;
unsigned long KBfree, KBtotal;
int prev_key;
static char recpartname[MAX_RECPARTNAME];
char ts_filename[MAX_PATH];
char mdb_filename[MAX_PATH];
char new_filename[MAX_PATH];
char freespace[30];
char totalspace[30];
unsigned long RECfile_MBsize;
unsigned long prev_files_size;
int divide_now;
int lowspace_now;
int emergency_stop;
int autostop;
time_t autostop_time;
int mode_teletext;
int mode_mplay;
int mode_internet;
int question;
int question_answer;
int EPG_progname_changed;
int EPG_progname_checks;
int current_divide;
int rec_history;
int telnetd_working;
int monthtab[] = {0,31,28,31,30,31,30,31,31,30,31,30,31};
int last_menulang = -1;
char *mess[MAX_MESS];
char *mess_buffer = NULL;
target targets[MAX_ALL_TARGETS];
int max_targets;
int max_const_targets;
static int cur_saving_mode;
static int panel_off;
int video_muted;
int esaving_restore;
int EPG_is_initialized = 0;
static unsigned EPG_channel[2];
static unsigned EPG_program[2];
static unsigned EPG_datetime[16];
static char EPG_progname[256];
char EPG_prev_progname[256];
unsigned short EPG_namebuffer[NAMEBUFSIZE];
unsigned EPG_namebuflen = NAMEBUFSIZE;
const char *OSD_SUBDIR		= "osd";
const char *ICON_TOOLBOX_HEADER = "toolbox_header";
const char *ICON_TOOLBOX_FOOTER = "toolbox_footer";
const char *ICON_TOOLBOX_FRAME	= "toolbox_frame";
const char *ICON_LEFT		= "icon_left";
const char *ICON_RIGHT		= "icon_right";
const char *ICON_UP		= "icon_up";
const char *ICON_DOWN		= "icon_down";
const char *ICON_UPDOWN		= "icon_updown";
const char *ICON_LEFTRIGHT	= "icon_leftright";
const char *ICON_ENTER		= "icon_enter";
const char *ICON_EXIT		= "icon_exit";
static SDL_Surface *screen;
int SDL_on;
int OSD_TTF_on;
unsigned *OSD_handle = NULL;
int (*ex_OSD_TTF_Init)(void);
int (*ex_OSD_TTF_WasInit)(void);
void (*ex_OSD_TTF_Quit)(void);
void *(*ex_OSD_TTF_OpenFont)(const char *, int);
void (*ex_OSD_TTF_CloseFont)(void *);
SDL_Surface *(*ex_OSD_TTF_RenderUTF8_Blended)(void *, const char *, SDL_Color);
int (*ex_OSD_TTF_SizeUTF8)(void *, const char *, int *, int *);
char *FONT_SHADOW		= "shadow.ttf";
void *ttf_font;
void *ttf_font18;
void *ttf_font20;
void *ttf_font26;
fontinfo fi_bold32[128] = {
{0x00000,20},{0x00080,20},{0x00100,20},{0x00180,15},{0x00200,15},{0x00280,17},{0x00300,17},{0x00380,18},
{0x00400,16},{0x00480,16},{0x00500,18},{0x00580,16},{0x00600,18},{0x00680,20},{0x00700,20},{0x00780,20},
{0x10000,20},{0x10080,19},{0x10100,18},{0x10180,18},{0x10200,13},{0x10280,19},{0x10300,18},{0x10380,17},
{0x10400,14},{0x10480,14},{0x10500,16},{0x10580,16},{0x10600,17},{0x10680,20},{0x10700,20},{0x10780,20},
{0x20000,20},{0x20080,12},{0x20100,20},{0x20180,20},{0x20200,20},{0x20280,20},{0x20300,20},{0x20380,14},
{0x20400,14},{0x20480,14},{0x20500,11},{0x20580,20},{0x20600,11},{0x20680,12},{0x20700,11},{0x20780,16},
{0x30000,20},{0x30080,17},{0x30100,19},{0x30180,20},{0x30200,20},{0x30280,20},{0x30300,20},{0x30380,18},
{0x30400,20},{0x30480,20},{0x30500,11},{0x30580,11},{0x30600,20},{0x30680,20},{0x30700,20},{0x30780,20},
{0x40000,20},{0x40080,20},{0x40100,20},{0x40180,20},{0x40200,20},{0x40280,15},{0x40300,15},{0x40380,20},
{0x40400,20},{0x40480,11},{0x40500,12},{0x40580,20},{0x40600,15},{0x40680,24},{0x40700,17},{0x40780,17},
{0x50000,17},{0x50080,17},{0x50100,17},{0x50180,17},{0x50200,17},{0x50280,17},{0x50300,20},{0x50380,28},
{0x50400,20},{0x50480,20},{0x50500,16},{0x50580,20},{0x50600,20},{0x50680,20},{0x50700,20},{0x50780,22},
{0x60000,20},{0x60080,19},{0x60100,20},{0x60180,19},{0x60200,20},{0x60280,19},{0x60300,13},{0x60380,20},
{0x60400,20},{0x60480,11},{0x60500,12},{0x60580,18},{0x60600,13},{0x60680,28},{0x60700,20},{0x60780,19},
{0x70000,20},{0x70080,19},{0x70100,14},{0x70180,17},{0x70200,14},{0x70280,20},{0x70300,19},{0x70380,25},
{0x70400,18},{0x70480,18},{0x70500,14},{0x70580,20},{0x70600,20},{0x70680,20},{0x70700,20},{0x70780,20}
};
fontinfo fi_dot18[128] = {
{0x0000,12},{0x000C,12},{0x0018,12},{0x0024,12},{0x0030,12},{0x003C,12},{0x0048,12},{0x0054,12},
{0x0060,12},{0x006C,12},{0x0078,12},{0x0084,12},{0x0090,12},{0x009C,12},{0x00A8,12},{0x00B4,12},
{0x0D80,12},{0x0D8C,12},{0x0D98,12},{0x0DA4,12},{0x0DB0,12},{0x0DBC,12},{0x0DC8,12},{0x0DD4,12},
{0x0DE0,12},{0x0DEC,12},{0x0DF8,12},{0x0E04,12},{0x0E10,12},{0x0E1C,12},{0x0E28,12},{0x0E34,12},
{0x1B00,12},{0x1B0C,12},{0x1B18,12},{0x1B24,12},{0x1B30,12},{0x1B3C,12},{0x1B48,12},{0x1B54,12},
{0x1B60,12},{0x1B6C,12},{0x1B78,12},{0x1B84,12},{0x1B90,12},{0x1B9C,12},{0x1BA8,12},{0x1BB4,12},
{0x2880,12},{0x288C,12},{0x2898,12},{0x28A4,12},{0x28B0,12},{0x28BC,12},{0x28C8,12},{0x28D4,12},
{0x28E0,12},{0x28EC,12},{0x28F8,12},{0x2904,12},{0x2910,12},{0x291C,12},{0x2928,12},{0x2934,12},
{0x3600,12},{0x360C,12},{0x3618,12},{0x3624,12},{0x3630,12},{0x363C,12},{0x3648,12},{0x3654,12},
{0x3660,12},{0x366C,12},{0x3678,12},{0x3684,12},{0x3690,12},{0x369C,12},{0x36A8,12},{0x36B4,12},
{0x4380,12},{0x438C,12},{0x4398,12},{0x43A4,12},{0x43B0,12},{0x43BC,12},{0x43C8,12},{0x43D4,12},
{0x43E0,12},{0x43EC,12},{0x43F8,12},{0x4404,12},{0x4410,12},{0x441C,12},{0x4428,12},{0x4434,12},
{0x5100,12},{0x510C,12},{0x5118,12},{0x5124,12},{0x5130,12},{0x513C,12},{0x5148,12},{0x5154,12},
{0x5160,12},{0x516C,12},{0x5178,12},{0x5184,12},{0x5190,12},{0x519C,12},{0x51A8,12},{0x51B4,12},
{0x5E80,12},{0x5E8C,12},{0x5E98,12},{0x5EA4,12},{0x5EB0,12},{0x5EBC,12},{0x5EC8,12},{0x5ED4,12},
{0x5EE0,12},{0x5EEC,12},{0x5EF8,12},{0x5F04,12},{0x5F10,12},{0x5F1C,12},{0x5F28,12},{0x5F34,12}
};
static SDL_Surface *font_bold32, *font_dot18, *font_lcd18;
const char *ICON_INFOPANE = "infopane";
const char *ICON_DESTUSB  = "destusb";
const char *ICON_DESTNET  = "destnet";
const char *ICON_DESTMEM  = "destmem";
const char *ICON_BAR_FREE = "bar_free";
const char *ICON_BAR_FULL = "bar_full";
int infopane_update;
int network_drive;
int internal_memory;
int free_level;
const char *ICON_WARNBOX_EXT	= "warnbox_ext";
const char *BUTTON_NONSEL	= "button";
const char *BUTTON_FOCUS	= "button_focus";
int question_but1x, question_but2x, question_but1y;
int warninfo_busy = 0;
const char *ICON_WARNBOX	= "warnbox";
const char *ICON_REC		= "icon_rec";
const char *ICON_STOPREC	= "icon_stoprec";
const char *ICON_STOP		= "icon_stop";
const char *ICON_NO		= "icon_no";
const char *ICON_SCRIPT		= "icon_script";
const char *ICON_RED		= "icon_red";
int totalfree_step;
int epg_step;
int divide_warned;
int lowspace_warned;
int monitoring_now;
int recwarnings_step;
int recwarned;
unsigned long epg_probes;
unsigned long next_probe;
char recpartname_best[MAX_RECPARTNAME];
char recpartname_prev[MAX_RECPARTNAME];
int setup_menu;
int selected_setup_item;
int setup_cursor;
int setup_offset;
int prev_selected_setup_item;
int prev_setup_cursor;
int prev_setup_offset;
char *SCRIPTS_DIR =		"scripts";
char *CONTROL_SCRIPT_FNAME = 	"/mtd_ram/_script.sh";
char *CONTROL_TAG_FNAME	= 	"/mtd_ram/_script.tag";
char *FAVCH_SCRIPT_FNAME =	"favch.sh";
char *SOURCE_SCRIPT_FNAME =	"source.sh";
int pause_state = PAUSE_OFF;
int button_is_pressed = 0;
int pressed_time;
int measuring_thread_working = 0;
int measuring_thread_ID;
pthread_t measuring_pthr;
int sustain_measuring_thread;
int measuring_key;

extern int _ZN4TCTv14PreferLanguageEv(int *);
extern int spIDp_Open(int, unsigned int *);
extern int spIDp_SetFreeze(unsigned int, int);
extern int spIDp_Close(unsigned int);

static int button_pressed(int key) __attribute__ ((noinline));

int run_telnetd(void)
{
   if (system("mount -t devpts devpts /dev/pts") == -1) return 1;
   if (system("telnetd") == -1) return 1;
   return 0;
}

int is_telnetd(void)
{
   FILE *f;
   char cmd[100];
   const char *logfile = "/mtd_ram/_is_td.log";

   sprintf(cmd, "ps|grep telnet[d] >%s", logfile);
   if (system(cmd) == -1) return 0;	
   f = fopen(logfile, "r");
   if (!f)
   {
      remove(logfile);
      return 0;
   }
   if (!fgets(cmd, sizeof(cmd), f))
   {
      fclose(f); remove(logfile);
      return 0;
   }
   fclose(f); remove(logfile);
   if (strstr(cmd, "telnetd")) return 1;
   return 0;
}

struct tm *localtime_from_UTC_system(int winter_time, int summer_time)
{
  time_t t;
  struct tm *tm;
  t = time(NULL);
  tm = localtime(&t); 
   
  if ((winter_time == 0) && (summer_time == 0)) return tm;
     
  int y = tm->tm_year; if (y<1900) y+=1900;
  int month = tm->tm_mon + 1;
  int day   = tm->tm_mday;
  int wday  = tm->tm_wday;
  int hour  = tm->tm_hour;

  int leap = ((y % 4) == 0)? 1:0; 
  if ((y % 100) == 0) 
     leap = ((y % 400) == 0)?1:0;
  monthtab[2] = 28 + leap;

  int zone = winter_time;
  if ((month) && (month<10)) zone = summer_time;
  else if ((month==3) || (month==10))
  {
     if ((day+7) > monthtab[month])
     {
        if (month==3) 
        {
           zone = summer_time;
           if ((wday==0) && (hour<1)) zone = winter_time;
        }
        if (month==10)
	{
	   if ((wday==0) && (hour<1)) zone = summer_time;
	   else if (wday + (monthtab[month] - day) >= 7) zone = summer_time;
	}
     }
     else
     {
        if (month==10) zone = summer_time;
     }
  }

  hour += zone;
  if (hour > 23)
  {
     hour -= 24;
     wday++;
     if (wday > 6) wday = 0;
     day++;
     if (day > monthtab[month])
     {
        day = 1;
        month++;
        if (month > 12)
        {
           month = 1;
           y++;
           tm->tm_year++;
        }
     }
  }
  else if (hour < 0)
  {
     hour += 24;
     wday--;
     if (wday < 0) wday = 6;
     day--;
     if (day < 1)
     {
        month--;
	if (month < 1)
	{
	   month = 12;
	   y--;
	   tm->tm_year--;
	}
	day = monthtab[month];
     }
  }
  
  tm->tm_wday = wday;
  tm->tm_mon  = month-1;
  tm->tm_mday = day;
  tm->tm_hour = hour;
  return tm;
}

int get_menu_language()
{
   int ret;
   ret = _ZN4TCTv14PreferLanguageEv((int *)g_AppTV);
   return ret;
}

char *lang_suffix(int menu_language)
{
   switch (menu_language)
   {
      case MENULANG_PL: return "PL";
      case MENULANG_DE: return "DE";
      case MENULANG_FR: return "FR";
      case MENULANG_BG: return "BG";
      case MENULANG_HR: return "HR";
      case MENULANG_CZ: return "CZ";
      case MENULANG_DK: return "DK";
      case MENULANG_NL: return "NL";
      case MENULANG_FI: return "FI";
      case MENULANG_GR: return "GR";
      case MENULANG_HU: return "HU";
      case MENULANG_IT: return "IT";
      case MENULANG_NO: return "NO";
      case MENULANG_PT: return "PT";
      case MENULANG_RO: return "RO";
      case MENULANG_RU: return "RU";
      case MENULANG_RS: return "RS";
      case MENULANG_SI: return "SI";
      case MENULANG_ES: return "ES";
      case MENULANG_SE: return "SE";
      case MENULANG_TR: return "TR";
      case MENULANG_EE: return "EE";
      case MENULANG_LV: return "LV";
      case MENULANG_LT: return "LT";

      default:		return "EN";
   }
}

void load_messages()
{
   int i, ln;
   char *pos, *stor;
   FILE *f;
   char s[256];
   char mess_fname[MAX_PATH];

   sprintf(mess_fname, "%smessages_%s.txt", config.path, lang_suffix(config.menu_language));
   if (access(mess_fname, F_OK) != 0) sprintf(mess_fname, "%smessages_%s.txt", config.path, lang_suffix(MENULANG_EN));

   struct stat fst;
   unsigned long mess_mem;
   if (stat(mess_fname, &fst) == 0) mess_mem = (unsigned long)fst.st_size;
   else mess_mem = 0L;

   for (i=0; i < MAX_MESS; i++) mess[i] = "?";

   if ((mess_mem < MIN_MESS_ALLOC) || (mess_mem > MAX_MESS_ALLOC)) return;

   mess_buffer = malloc(mess_mem - 2);
   if (mess_buffer)
   {
      stor = mess_buffer;

      f = fopen(mess_fname, "rt");
      if (f)
      {
	 while (!feof(f))
	 {
	    if (fgets(s, sizeof(s)-1, f) == s)
	    {
	       ln = strlen(s);
	       if (ln > 0) for (i=0; i < ln; i++) if ((s[i]==CR) || (s[i]==LF)) s[i]=0;
	       ln = strlen(s);
	       if (ln > 0)
	          if (s[0] != ';')
		  {
		     pos = strstr(s, "=");
		     if (pos > s)
		     {
			*pos = 0;
			i = atoi(s);
			if ((i >= 0) && (i < MAX_MESS))
			{
			   strcpy(stor, pos+1);
			   mess[i] = stor;
			   stor += strlen(stor);
			   *(stor++) = 0;
			}
		     }
		  }
	    }
	 }
	 fclose(f);
      }
   }

   last_menulang = config.menu_language;
}

void unload_messages()
{
   if (mess_buffer) { free(mess_buffer); mess_buffer = NULL; }
}

void debug(char *text)
{
   FILE *f;
   char s[256];

   sprintf(s, "%s\n", text);
   f = fopen(DEBUG_LOG, "at");
   if (f) { fputs(s, f); fflush(f); fclose(f); }
}

int no_function(unsigned *handle, char *function)
{
   char s[128];
   sprintf(s, "Function [%s] not found:", function);
   debug(s);
   debug(dlerror());
   if (handle) dlclose(handle);
   return 0;
}

int strequal(char *s1, char *s2)
{
   int ln1, ln2;

   ln1 = strlen(s1);
   ln2 = strlen(s2);
   if (ln1 != ln2) return 0;
   if (ln1 == 0) return 0;
   if (strstr(s1, s2)) return 1;
   return 0;
}

int is_NTFS(char *path)
{
   if (statfs(path, &fs) == 0)
   {
      if (fs.f_type == FS_NTFS) return 1;
   }
   return 0;
}

void parse_usb_log()
{
   char s[MAX_LINE+1];
   char *p;
   FILE *f;
   int i, ln, partition;
   target temp_target;

   memset(&temp_target, 0, sizeof(temp_target)); temp_target.id = 0; temp_target.name[0] = 0;
   max_targets = max_const_targets;
   partition = 0;

   f = fopen("/dtv/usb/log", "rt");
   if (f)
   {
      while(!feof(f))
      {
	 if (fgets(s, MAX_LINE, f))
	 {
	    ln = strlen(s);
	    if (ln > 0)
	    {
	       for (i=0; i<ln; i++) if ((s[i] == 0x0D) || (s[i] == 0x0A)) s[i] = 0;
	       ln = strlen(s);
	       if (ln > 0)
	       {
		  if (s[0] == '[')
		  {
		     s[MAX_TARGET_NAME] = 0;
		     ln = strlen(s);
		     if (ln > 1)
		     {
			for (i=1; i<ln; i++) if (s[i] == ']') s[i] = 0;
			ln = strlen(s+1);
			if (ln > 0)
			{
			   memset(&temp_target, 0, sizeof(temp_target));
			   strcpy(temp_target.name, s+1);
			   temp_target.vendor[0] = 0;
			   temp_target.product[0] = 0;
			   temp_target.serial[0] = 0;
			   temp_target.path[0] = 0;
			   temp_target.usesintmem = 0;
			   temp_target.not_allowed = 0;
			   partition = 1;
			}
		     }
		  }
		  else
		  {
		     p = strstr(s, "Vendor : "); if (p==s) { s[sizeof("Vendor :") + MAX_TARGET_VENDOR] = 0; strcpy(temp_target.vendor, p + sizeof("Vendor :")); }
		     p = strstr(s, "Product : "); if (p==s) { s[sizeof("Product :") + MAX_TARGET_PRODUCT] = 0; strcpy(temp_target.product, p + sizeof("Product :")); }
		     p = strstr(s, "Serial : "); if (p==s) { s[sizeof("Serial :") + MAX_TARGET_SERIAL] = 0; strcpy(temp_target.serial, p + sizeof("Serial :")); }
		     p = strstr(s, "MountDir : "); if (p==s)
		     {
			if (temp_target.path[0])
			{
			   partition++;
			   temp_target.not_allowed = 0;
			}
			s[sizeof("MountDir :") + MAX_TARGET_PATH] = 0; strcpy(temp_target.path, p + sizeof("MountDir :"));
			ln = strlen(temp_target.vendor); while (ln > 0) { if (temp_target.vendor[ln-1] == ' ') temp_target.vendor[ln-1] = 0; else ln=0; ln--; }
			ln = strlen(temp_target.product); while (ln > 0) { if (temp_target.product[ln-1] == ' ') temp_target.product[ln-1] = 0; else ln=0; ln--; }
			if (partition > 1)
			{
			   sprintf(temp_target.product + strlen(temp_target.product), " %d", partition);
			   if ((partition == 2) && (max_targets > max_const_targets)) strcat(targets[max_targets-1].product, " 1");
			}
			if ((max_targets < MAX_ALL_TARGETS) &&
			    (temp_target.name[0]) &&
			    (temp_target.path[0]) &&
			    (partition >= 1))
			{
			   if (is_NTFS(temp_target.path)) temp_target.not_allowed = 1;
			   temp_target.id = max_targets;
			   targets[max_targets++] = temp_target;
			}
		     }
		  }
	       }
	    }
	 }
      }
      fclose(f);
   }
}

void init_targets()
{
   target AUTO_TARGET = AUTO_TARGET_DEF;
   target INTERNAL_TARGET = INTERNAL_TARGET_DEF;
   max_targets = 0;
   targets[max_targets++] = AUTO_TARGET;
   targets[max_targets++] = INTERNAL_TARGET;
   max_const_targets = max_targets;
   parse_usb_log();
}

int find_target(target *targ)
{
   int i;
   int ret = 0;

   if (targ->id < max_const_targets) return targ->id;

   if (max_targets >= max_const_targets)
      for (i=max_const_targets; i < max_targets; i++)
      {
	  if (strequal(targets[i].product, targ->product)) ret = i;
	  else if (strequal(targets[i].path, targ->path))
		  { if (ret==0) ret = i; }
	  else if (strequal(targets[i].name, targ->name))
		  { if (ret==0) ret = i; }
      }
   return ret;
}

void refresh_config_targets()
{
   config.target1 = targets[find_target(&config.target1)];
   config.target2 = targets[find_target(&config.target2)];
}

void store_config_target1()
{
   int id = config.target1.id;
   config.target1 = targets[id];
}
void store_config_target2()
{
   int id = config.target2.id;
   config.target2 = targets[id];
}

static void esaving_togglevideo()
{
   static int mflag = 1;
   _ZN8TCWindow9MuteVideoEb(&g_AppWindow, mflag);
   _ZN8TCWindow9MuteVideoEb(&g_AppWindowPIP, mflag);

   video_muted = mflag?1:0;
   mflag = !mflag;
}

static unsigned int esaving_get_option(int option)
{
	unsigned int varr[2];
	unsigned int *rval=varr;
	int retv;

	memset(varr, 0, sizeof(varr));
	retv = _ZN4TCTv9GetOptionEiPi(g_AppTV, option, rval);
	return (*rval);
}

static int esaving_set_option(int option, int mode)
{
	int retv;

	retv = _ZN4TCTv9SetOptionEii(g_AppTV, option, mode);
	return (retv);
}

static void esaving_panel_toggle()
{
   int ptevent[6];

   if (panel_off)
   {
      panel_off = 0;
      esaving_restore = 1;
      esaving_set_option(TCTV_OPTION_ESAVING, cur_saving_mode);
   }
   else
   {
      cur_saving_mode = esaving_get_option(TCTV_OPTION_ESAVING);
      esaving_set_option(TCTV_OPTION_ESAVING, ESAVING_PANEL_OFF);
      memset(ptevent, 0, sizeof(ptevent));
      ptevent[3] = 0x432;
      ptevent[4] = 1;
      ptevent[5] = 0;

      if (config.model == MODEL_CHL7DEUC) _ZN4CApp9SendEventEP7PTEvent15DTV_APPLICATIONb(ptevent + 1, 2, 0);
      else _ZN4CApp9SendEventEl15DTV_APPLICATIONb(ptevent + 1, 2, 0);

      panel_off = 1;
   }
}

void esaving_panel_off()
{
   if (!panel_off) esaving_panel_toggle();
}

void esaving_panel_on()
{
   if (panel_off) esaving_panel_toggle();
}

void esaving_init()
{
   panel_off = 0;
   video_muted = 0;
   esaving_restore = 0;
   cur_saving_mode = esaving_get_option(TCTV_OPTION_ESAVING);
}

static int EPG_unicode2ascii( unsigned short *unicode, char *ascii ) __attribute__ ((noinline));
static int EPG_unicode2ascii( unsigned short *unicode, char *ascii )
{
   unsigned short *psrc;
   char *pdst, *pconv;
   int i, found;

   psrc = unicode;
   pdst = ascii;

   while (*psrc)
   {
      found=0;
      if (config.EPG_transcode==1)
      {
	 i=0;
	 while ((unicode2ascii[i].unicode != 0x000) && !found)
	 {
	    if (*psrc == unicode2ascii[i].unicode)
	    {
	       pconv = unicode2ascii[i].ascii;
	       while (*pconv) *(pdst++) = *(pconv++);
	       found = 1;
	    }
	    i++;
	 }
      }
      if (!found) *(pdst++) = *psrc >> 8;
      psrc++;
   }
   *pdst = 0;

   while (--pdst >= ascii)
   {
      if (!(  ((*pdst>='0') && (*pdst<='9')) ||
	      ((*pdst>='a') && (*pdst<='z')) ||
	      ((*pdst>='A') && (*pdst<='Z')) ||
	       (*pdst=='-') || (*pdst=='!') || (*pdst=='$') || (*pdst=='@') ||
	       (*pdst=='&') || (*pdst=='(') || (*pdst==')')
	     )) *pdst = REPLACEMENT;
   }

   return 0;
}

static int unicode2utf8(unsigned short *unicode, char *utf8, unsigned utf8size, int remove_bad) __attribute__ ((noinline));
static int unicode2utf8(unsigned short *unicode, char *utf8, unsigned utf8size, int remove_bad)
{
	unsigned short uchar;
	unsigned cchar;
	int i, j;

	*utf8 = '\0';
	for (cchar = 0; (uchar = ntohs(unicode[cchar])) != 0; cchar++)
	{
	    if (utf8size <= UTF8_TABLE1_SIZE) break;

	    for (i = 0; i < UTF8_TABLE1_SIZE; i++)
		if (uchar <= utf8_table1[i]) break;
	    utf8 += i;
	    utf8[1] = '\0';
	    for (j = i; j > 0; j--)
	    {
		*utf8-- = 0x80 | (uchar & 0x3F);
		uchar >>= 6;
	    }
	    *utf8 = utf8_table2[i] | uchar;
	    if ((remove_bad) && (i == 0))
	    {
	       char *bad = "/\"?:<>*\\|";
	       if (strchr(bad, *utf8) != NULL) *utf8 = REPLACEMENT;
	    }

	    i++;
	    utf8 += i;
	    utf8size -= i;
	}
	*utf8 = '\0';

	return (0);
}

void EPG_init()
{
	EPG_is_initialized = 1;
	EPG_progname_changed = 0;
	EPG_progname_checks = 0;
	memset(EPG_datetime, 0, 16*sizeof(unsigned));
	_ZN9TCChannelC1Ev(EPG_channel);
	_ZN9TCProgramC1Ev(EPG_program);
	mdb_filename[0] = 0;
}

void EPG_get_current_time()
{
	_ZN4TCTv17GetSystemInfoTimeEPmi(*g_AppTV, EPG_datetime, 0);
}

void EPG_get_structured_current_time()
{
	_ZN10CTimerUtil10GetSITTimeEPN6PCTime6PTTimeEi(EPG_datetime, 0);
}

static int EPG_get_current_channel()
{
	if ( _ZN8CAppUtil12GetCurrentChEP9TCChanneli(EPG_channel, 0) ) return 1;
	return 0;
}

static int EPG_get_present_program()
{
	if ( _ZN4TCTv17GetPresentProgramEP9TCChanneljP9TCProgram(*g_AppTV, EPG_channel, EPG_datetime, EPG_program) ) return 1;
	return 0;
}

void EPG_get_program_title()
{
	if (EPG_progname_changed == 0) strcpy(EPG_prev_progname, EPG_progname);

	memset(EPG_namebuffer, 0, NAMEBUFSIZE);
	EPG_namebuflen = NAMEBUFSIZE;
	_ZNK9TCProgram8GetTitleEPcPj(EPG_program, EPG_namebuffer, &EPG_namebuflen);
	if (config.EPG_transcode == EPG_TRANS_UTF8)
	   unicode2utf8(EPG_namebuffer, EPG_progname, sizeof(EPG_progname), 1);
	else
	   EPG_unicode2ascii(EPG_namebuffer, EPG_progname);

	if (EPG_progname_changed < EPG_PROGNAME_CHANGED_LIMIT)
	{
	   if (strlen(EPG_prev_progname) != strlen(EPG_progname))	EPG_progname_changed++;
	   else if (!strstr(EPG_prev_progname, EPG_progname))		EPG_progname_changed++;

	   unsigned long dur = (unsigned long)(time(NULL) - ((part_number > 0)?rec_partstarttime:rec_starttime));
	   if (dur < ((EPG_FIRST_PERIOD + 2*EPG_PERIOD) * MONITORING_INTERVAL / 1000)) { EPG_progname_changed = 0; EPG_progname_checks = 0; }

	   if ((EPG_progname_changed > 0) && (EPG_progname_changed < EPG_PROGNAME_CHANGED_LIMIT))
	   {
	      EPG_progname_checks++;
	      if (EPG_progname_checks > EPG_PROGNAME_CHANGED_LIMIT*2) { EPG_progname_changed = 0; EPG_progname_checks = 0; }
	   }
	}
}

int EPG_get_data()
{
	struct tm *tm;
	int y;

	if (!EPG_is_initialized) EPG_init();

	EPG_get_current_time();
	if (EPG_get_current_channel())
	   if (EPG_get_present_program())
	   {
	      EPG_get_program_title();
	      EPG_get_structured_current_time();
	      EPG_progname[MAX_PROGNAME] = 0;
	      if (config.timesource == TIMESOURCE_NTP)
	      {
		 tm = localtime_from_UTC_system(timezones[config.timezone].winter_time, timezones[config.timezone].summer_time);
		 y = tm->tm_year; if (y<1900) y+=1900;
		 if ((y >= MIN_YEAR) && (y <= MAX_YEAR))
		 {
		    EPG_year = y; EPG_month = tm->tm_mon+1; EPG_day = tm->tm_mday;
		    EPG_hour = tm->tm_hour; EPG_minute = tm->tm_min;
		 }
	      }

	      if ((EPG_year >= MIN_YEAR) && (EPG_year <= MAX_YEAR))
	      {
	         sprintf(recpartname, "%4d-%02d-%02d_%02d-%02d_%s_",
			EPG_year, EPG_month, EPG_day,
			EPG_hour, EPG_minute,
			EPG_progname);
	         return 1;
	      }
	   }

	tm = localtime_from_UTC_system(timezones[config.timezone].winter_time, timezones[config.timezone].summer_time);
	y = tm->tm_year; if (y<1900) y+=1900;
	if ((y >= MIN_YEAR) && (y <= MAX_YEAR))
	    sprintf(recpartname, "%4d-%02d-%02d_%02d-%02d_", y, tm->tm_mon+1, tm->tm_mday, tm->tm_hour, tm->tm_min);
	else strcpy(recpartname, "_");

	return 0;
}

int OSD_init_ttf_helper()
{
   char s[MAX_PATH];

   if (OSD_TTF_on) return 1;

   sprintf(s, "%s%s", config.path, HELPER_MODULE);

   if ( (OSD_handle = dlopen(s, RTLD_LAZY)) == NULL ) return 0;

   ex_OSD_TTF_Init = dlsym(OSD_handle, "OSD_TTF_Init");
   if (ex_OSD_TTF_Init == NULL) return no_function(OSD_handle, "OSD_TTF_Init");

   ex_OSD_TTF_WasInit = dlsym(OSD_handle, "OSD_TTF_WasInit");
   if (ex_OSD_TTF_WasInit == NULL) return no_function(OSD_handle, "OSD_TTF_WasInit");

   ex_OSD_TTF_Quit = dlsym(OSD_handle, "OSD_TTF_Quit");
   if (ex_OSD_TTF_Quit == NULL) return no_function(OSD_handle, "OSD_TTF_Quit");

   ex_OSD_TTF_OpenFont = dlsym(OSD_handle, "OSD_TTF_OpenFont");
   if (ex_OSD_TTF_OpenFont == NULL) return no_function(OSD_handle, "OSD_TTF_OpenFont");

   ex_OSD_TTF_CloseFont = dlsym(OSD_handle, "OSD_TTF_CloseFont");
   if (ex_OSD_TTF_CloseFont == NULL) return no_function(OSD_handle, "OSD_TTF_CloseFont");

   ex_OSD_TTF_RenderUTF8_Blended = dlsym(OSD_handle, "OSD_TTF_RenderUTF8_Blended");
   if (ex_OSD_TTF_RenderUTF8_Blended == NULL) return no_function(OSD_handle, "OSD_TTF_RenderUTF8_Blended");

   ex_OSD_TTF_SizeUTF8 = dlsym(OSD_handle, "OSD_TTF_SizeUTF8");
   if (ex_OSD_TTF_SizeUTF8 == NULL) return no_function(OSD_handle, "OSD_TTF_SizeUTF8");

   OSD_TTF_on = 1;
   return 1;
}

void *open_TTF_font(char *font_fname, int size)
{
      char s[MAX_PATH];
      sprintf(s, "%s/%s", FONTS_SUBDIR, font_fname);
      if (access(s, F_OK) != 0) sprintf(s, "%s%s/%s", config.path, LOCAL_FONTS_SUBDIR, font_fname);
      return OSD_TTF_OpenFont(s, size);
}

void init_OSD_fonts()
{
   font_bold32  = NULL;
   font_dot18   = NULL;
   font_lcd18	= NULL;

   if (ttf_font18 == NULL) ttf_font18 = open_TTF_font(FONT_SHADOW, 18);
   if (ttf_font20 == NULL) ttf_font20 = open_TTF_font(FONT_SHADOW, 20);
   if (ttf_font26 == NULL) ttf_font26 = open_TTF_font(FONT_SHADOW, 26);
}

char national_char(char *ch2)
{
   unsigned int utf;

   char ch = SPC;

   utf = (((unsigned char)ch2[0]<<8) + (unsigned char)ch2[1]) & 0xFFFF;

   switch(utf)
   {
	case 0xC485: ch = apl; break;
	case 0xC487: ch = cpl; break;
	case 0xC499: ch = epl; break;
	case 0xC582: ch = lpl; break;
	case 0xC584: ch = npl; break;
	case 0xC3B3: ch = opl; break;
	case 0xC59B: ch = spl; break;
	case 0xC5BA: ch = xpl; break;
	case 0xC5BC: ch = zpl; break;
	case 0xC484: ch = Apl; break;
	case 0xC486: ch = Cpl; break;
	case 0xC498: ch = Epl; break;
	case 0xC581: ch = Lpl; break;
	case 0xC583: ch = Npl; break;
	case 0xC393: ch = Opl; break;
	case 0xC59A: ch = Spl; break;
	case 0xC5B9: ch = Xpl; break;
	case 0xC5BB: ch = Zpl; break;
	case 0xC3A4: ch = ade; break;
	case 0xC3B6: ch = ode; break;
	case 0xC3BC: ch = ude; break;
	case 0xC39F: ch = ssde; break;
	case 0xC384: ch = Ade; break;
	case 0xC396: ch = Ode; break;
	case 0xC39C: ch = Ude; break;
	case 0xC3A8: ch = eGfr; break;
	case 0xC3A9: ch = eAfr; break;
	case 0xC3AA: ch = eCfr; break;
	default: ;
   }
   return ch;
}

void show_text(int font_num, int xpos, int ypos, char *text, int color)
{
   int i, ln;
   int x, y;
   unsigned char Y;
   char s[256];

   unsigned char Rc, Gc, Bc;
   Rc = colortable[color].r;
   Gc = colortable[color].g;
   Bc = colortable[color].b;

   ln = strlen(text);

   if ((xpos<0) || (ypos<0) || (ln==0)) return;

   if (screen)
   {
      if ((font_num == SHADOW18) || (font_num == SHADOW20) || (font_num == SHADOW26))
      {
	 if (font_num == SHADOW18) ttf_font = ttf_font18;
	 else
	 if (font_num == SHADOW20) ttf_font = ttf_font20;
	 else
	 if (font_num == SHADOW26) ttf_font = ttf_font26;

	 SDL_Surface *ttf_text;
	 if (ttf_font)
	 {
	    SDL_Rect ttf_rect = {xpos, ypos, SCREEN_W-xpos, SCREEN_H-ypos};
	    ttf_text = OSD_TTF_RenderUTF8_Blended(ttf_font, text, colortable[color]);
	    SDL_BlitSurface(ttf_text, NULL, screen, &ttf_rect);
	    SDL_FreeSurface(ttf_text);
	    return;
	 }
      }

      sprintf(s, "%s%s/font%d.png", config.path, LOCAL_FONTS_SUBDIR, font_num);

      SDL_Surface *font_surface;
      fontinfo* fi;
      int fi_height;

      switch (font_num)
      {
	 case DOT18:
		font_surface = font_dot18;
		fi = fi_dot18;
		fi_height = FI_DOT18_HEIGHT;
		break;
	 case LCD18:
		font_surface = font_lcd18;
		fi = fi_dot18;
		fi_height = FI_DOT18_HEIGHT;
		break;
	 default:
		return;
      }
   
      if (!font_surface) font_surface = IMG_Load(s);
      if (font_surface)
      {
	 switch (font_num)
	 {
	    case DOT18:   font_dot18   = font_surface; break;
	    case LCD18:   font_lcd18   = font_surface; break;
	    default: return;
	 }

	 SDL_LockSurface(screen);
	 SDL_LockSurface(font_surface);
	 unsigned char *scr, *scr_base = (unsigned char *)screen->pixels;
	 unsigned char *fnt, *fnt_base = (unsigned char *)font_surface->pixels;

	 int xpos_max = screen->pitch / (SCREEN_BPP/8);
	 int ypos_max = screen->h;

	 for (i=0; i<ln; i++)
         {
	     unsigned char ch = text[i];
	     if (ch > 127) ch = national_char(text+(i++));
	     if (ch > 127) ch -= 128;
	     if (xpos > xpos_max - fi[ch].width)
	     {
		xpos = 0;
		ypos += fi_height;
	     }
	     if (ypos > ypos_max) ypos = 0;

	     fnt = fnt_base + fi[ch].offset;
	     scr = scr_base + ypos * screen->pitch + xpos * (SCREEN_BPP/8);

	     if (ch != ' ')
	     for (y = 0; y < fi_height; y++)
             {
		 int wid = fi[ch].width;
                 for (x = 0; x < wid; x++)
	         {
		     Y = *(fnt++);
		     if (Y == 0) scr+= (SCREEN_BPP/8);
		     else
		     {
		        *(scr++) = (unsigned char)(Y * Bc/255);
		        *(scr++) = (unsigned char)(Y * Gc/255);
		        *(scr++) = (unsigned char)(Y * Rc/255);
		        *(scr++) = 255;
		     }
	         }
		 fnt += font_surface->pitch - fi[ch].width;
                 scr += screen->pitch - fi[ch].width * (SCREEN_BPP/8);
             }
	     xpos += fi[ch].width;
	 }

	 SDL_UnlockSurface(font_surface);
	 SDL_UnlockSurface(screen);
      }
   }
}

int get_text_width(int font_num, char *text)
{
   int i, ln, h, w;

   ln = strlen(text);
   if (ln==0) return 0;

   if ((font_num == SHADOW18) || (font_num == SHADOW20) || (font_num == SHADOW26))
   {
      if (font_num == SHADOW18) ttf_font = ttf_font18;
      else
      if (font_num == SHADOW20) ttf_font = ttf_font20;
      else
      if (font_num == SHADOW26) ttf_font = ttf_font26;

      if (ttf_font)
      {
	 i = OSD_TTF_SizeUTF8(ttf_font, text, &w, &h);
	 if (i == 0) return w;
      }
   }

   fontinfo* fi;
   switch (font_num)
   {
      case DOT18:   fi = fi_dot18;   break;
      case LCD18:   fi = fi_lcd18;   break;
      default: return SCREEN_W+1;
   }

   w = 0;
   for (i=0; i<ln; i++)
   {
       unsigned char ch = text[i];
       if (ch > 127) ch = national_char(text+(i++));
       if (ch > 127) ch -= 128;
       w += fi[ch].width;
   }
   return w;
}

void show_bold_text(int xpos, int ypos, char *text)
{
   int i, ln;
   int x, y;
   unsigned char R, G, B, A;
   char s[256];

   ln = strlen(text);

   if ((xpos<0) || (ypos<0) || (ln==0)) return;

   if (screen)
   {
      sprintf(s, "%s%s/font32.png", config.path, LOCAL_FONTS_SUBDIR);
   
      if (!font_bold32) font_bold32 = IMG_Load(s);
      if (font_bold32)
      {
	 SDL_LockSurface(screen);
	 SDL_LockSurface(font_bold32);
	 unsigned char *scr, *scr_base = (unsigned char *)screen->pixels;
	 unsigned char *fnt, *fnt_base = (unsigned char *)font_bold32->pixels;
	 int fnt_bpp = font_bold32->format->BytesPerPixel;
	 int xpos_max = screen->pitch / (SCREEN_BPP/8);
	 int ypos_max = screen->h;

	 R = 8; G = 8; B = 8; A = 128;

	 for (i=0; i<ln; i++)
         {
	     unsigned char ch = text[i];
	     if (ch > 127) ch = national_char(text+(i++));
	     if (ch > 127) ch -= 128;
	     if (xpos > xpos_max - fi_bold32[ch].width)
	     {
		xpos = 0;
		ypos += FI_BOLD32_HEIGHT;
	     }
	     if (ypos > ypos_max) ypos = 0;

	     fnt = fnt_base + fi_bold32[ch].offset;
	     scr = scr_base + ypos * screen->pitch + xpos * (SCREEN_BPP/8);

	     if (ch != ' ')
	     for (y = -3; y < FI_BOLD32_HEIGHT; y++)
             {
		 int wid = fi_bold32[ch].width;
                 for (x = 0; x < wid; x++)
	         {
		     if (y >= 0)
		     {
		        R = *(fnt++);
		        G = *(fnt++);
		        B = *(fnt++);
		        if (fnt_bpp == 4) A = *(fnt++);
		     }

		     *(scr++) = B;
		     *(scr++) = G;
		     *(scr++) = R;
		     *(scr++) = A * config.OSD/100 * config.alpha/100;
	         }
		 if (y >= 0) fnt += font_bold32->pitch - fi_bold32[ch].width * fnt_bpp;
                 scr += screen->pitch - fi_bold32[ch].width * (SCREEN_BPP/8);
             }

	     xpos += fi_bold32[ch].width;
	 }
	 SDL_UnlockSurface(font_bold32);
	 SDL_UnlockSurface(screen);
      }
   }
}

void init_OSD()
{
   screen = NULL;
   SDL_on = 0;
   OSD_TTF_on = 0;
}

void close_OSD()
{
   if (is_recording) return;

   SDL_Quit();
   SDL_on = 0;
   screen = NULL;
}

void init_video_mode()
{
   if (!SDL_on)
   {
      SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER);
      SDL_on = 1;
   }

   if (!OSD_init_ttf_helper()) { close_OSD(); return; }
   if (!OSD_TTF_WasInit()) if (OSD_TTF_Init() != 0) return;
   init_OSD_fonts();
   if (screen) return;
   screen = SDL_SetVideoMode(SCREEN_W, SCREEN_H, SCREEN_BPP, SCREEN_FLAGS);
}

void close_video_mode()
{
   if (font_bold32)  { SDL_FreeSurface(font_bold32);  font_bold32  = NULL; }
   if (font_dot18)   { SDL_FreeSurface(font_dot18);   font_dot18   = NULL; }
   if (ttf_font18) { OSD_TTF_CloseFont(ttf_font18); ttf_font18 = NULL; }
   if (ttf_font20) { OSD_TTF_CloseFont(ttf_font20); ttf_font20 = NULL; }
   if (ttf_font26) { OSD_TTF_CloseFont(ttf_font26); ttf_font26 = NULL; }
   if (OSD_TTF_WasInit()) OSD_TTF_Quit();
   close_OSD();
}

void OSD_cls()
{
   unsigned char *scr;
   int y;

   if (screen)
   {
      SDL_LockSurface(screen);
      scr = screen->pixels;
      for (y = 0; y < SCREEN_H; y++)
      {
	  memset(scr, 0, SCREEN_W * SCREEN_BPP/8);
	  scr += screen->pitch;
      }
      SDL_UnlockSurface(screen);
      SDL_Flip(screen);
   }
}

int draw_bitmap(int xpos, int ypos, const char *pngname, int enlarge)
{
   static SDL_Surface *bitmap;
   int x, y;
   unsigned char R, G, B, A = 255;
   char s[MAX_PATH];

   if (screen)
   {
      sprintf(s, "%s%s/%s.png", config.path, OSD_SUBDIR, pngname);
   
      bitmap = IMG_Load(s);
      if (bitmap)
      {
	    SDL_LockSurface(screen);
	    SDL_LockSurface(bitmap);
	    unsigned char *scr = (unsigned char *)screen->pixels;
	    unsigned char *bmp = (unsigned char *)bitmap->pixels;
	    int bmp_bpp = bitmap->format->BytesPerPixel;
	    int enlarge_h = bitmap->h / 2;

	    scr += ypos * screen->pitch + xpos * (SCREEN_BPP/8);
	    for (y = 0; y < bitmap->h; y++)
            {
                for (x = 0; x < bitmap->w; x++)
		{
		    R = *(bmp++);
		    G = *(bmp++);
		    B = *(bmp++);
		    if (bmp_bpp == 4) A = *(bmp++) * config.OSD / 100;
		    
		    if ((enlarge == -1) && (R+G+B == 0)) scr += (SCREEN_BPP/8);
		    else
		    {
		       *(scr++) = B;
		       *(scr++) = G;
		       *(scr++) = R;
		       *(scr++) = A;
		    }
		}
                scr += screen->pitch - bitmap->w * (SCREEN_BPP/8);

	        if ((y == enlarge_h) && (enlarge>0)) { y--; bmp -= bitmap->w * bmp_bpp; enlarge--; }

		bmp += bitmap->pitch - bitmap->w * bmp_bpp;
            }
	    SDL_UnlockSurface(bitmap);
	    SDL_UnlockSurface(screen);

	    SDL_FreeSurface(bitmap); bitmap = NULL;

            return 1;
      }
   }
   return 0;
}

void mix_png(SDL_Surface *bkgpic, const char *pngname, int mix_x, int mix_y, int skip_lines)
{
   int x, y, w, h, offset, bpp;
   unsigned char R, G, B, A, invA;
   unsigned char *bkg, *mix;
   char s[MAX_PATH];
   SDL_Surface *mixpic;

   sprintf(s, "%s%s/%s.png", config.path, OSD_SUBDIR, pngname);

   offset = (mix_y + skip_lines) * bkgpic->pitch + mix_x * bkgpic->format->BytesPerPixel;

   mixpic = IMG_Load(s);
   if (mixpic && bkgpic)
   {
      SDL_LockSurface(bkgpic);
      SDL_LockSurface(mixpic);

      h = mixpic->h;
      w = mixpic->w;
      mix = (unsigned char *)mixpic->pixels + skip_lines * mixpic->pitch;
      bpp = mixpic->format->BytesPerPixel;

      bkg = (unsigned char *)bkgpic->pixels + offset;
      A = 255;
      for (y=skip_lines; y<h; y++)
      {
          for (x=0; x<w; x++)
	  {
	      R = *(mix++);
	      G = *(mix++);
	      B = *(mix++);
	      if (bpp==4) A = *(mix++);

	      if (R+G+B == 0) bkg += 4;
	      else
	      {
		invA = 255-A;
	        *(bkg) = (unsigned char)((R * A + (*bkg) * invA) >> 9); bkg++;
	        *(bkg) = (unsigned char)((G * A + (*bkg) * invA) >> 9); bkg++;
	        *(bkg) = (unsigned char)((B * A + (*bkg) * invA) >> 9); bkg++;
	        bkg++;
	      }
	  }
	  bkg += bkgpic->pitch - w * 4;
      }

      SDL_UnlockSurface(mixpic);
      SDL_UnlockSurface(bkgpic);
   }
}

void change_alpha(int x, int y, int w, int h, unsigned char A)
{
   int i;
   unsigned char *scr;

   if (screen)
   {
      SDL_LockSurface(screen);
      scr = screen->pixels + y * screen->pitch + x * (SCREEN_BPP/8) + 3;

      while (h > 0)
      {
	 for (i=0; i < w; i++) { *(scr) = A; scr+=(SCREEN_BPP/8); }
	 scr += screen->pitch - w * (SCREEN_BPP/8);
	 h--;
      }
      SDL_UnlockSurface(screen);
   }
}

void mix_dest_icon(SDL_Surface *icon)
{
   mix_png(icon, internal_memory?ICON_DESTMEM:(network_drive?ICON_DESTNET:ICON_DESTUSB), 20, 24, 0);
}

void infopane_update_thread()
{
   int x, y, w, h, xpos, ypos, min_left;
   unsigned char R, G, B, A = 255;
   unsigned char *scr, *ico;
   SDL_Surface *icon;
   char s[MAX_PATH];
   char s2[100];

   int infofont = infopanefonts[config.infopanefont].font;
   int infowidth = infopanefonts[config.infopanefont].max_width;

   config.menu_language = get_menu_language();
   if (config.menu_language != last_menulang) load_messages();

   if (!infopane_update) return;

   if (screen)
   {
      sprintf(s, "%s%s/%s.png", config.path, OSD_SUBDIR, ICON_INFOPANE);
   
      icon = IMG_Load(s);
      if (icon)
      {
	 h = icon->h;
	 w = icon->w;

	 mix_dest_icon(icon);

	 if (free_level != -1)
	 {
	    mix_png(icon, ICON_BAR_FREE, 44, 80, 0);
	    mix_png(icon, ICON_BAR_FULL, 44, 80, free_level);
	 }

	 SDL_LockSurface(icon);
	 ico = (unsigned char *)icon->pixels;
	 xpos = (screen->w - w) >> 1;
	 ypos = screen->h - h;

	 SDL_LockSurface(screen);
	 scr = screen->pixels + ypos * screen->pitch + xpos * (SCREEN_BPP/8);
	 for (y = 0; y < h; y++)
         {
             for (x = 0; x < w; x++)
	     {
		 R = *(ico++);
		 G = *(ico++);
		 B = *(ico++);
		 A = *(ico++);
		    
		 *(scr++) = B;
		 *(scr++) = G;
		 *(scr++) = R;
		 *(scr++) = A * config.alpha/100 * config.OSD/100;
	     }
             scr += screen->pitch - w * (SCREEN_BPP/8);
         }
	 SDL_UnlockSurface(screen);
	    
	 SDL_UnlockSurface(icon);
	 SDL_FreeSurface(icon); icon = NULL;

	 int dur = (int)(time(NULL) - rec_starttime);
	 int hh = dur / 3600;
	 int mm = (dur - hh*3600) / 60;
	 int ss = dur - hh*3600 - mm*60;
	 sprintf(s, "%d:%02d:%02d", hh, mm, ss);
	 show_text(infofont, 790 - get_text_width(infofont, s), INFOPANE_TOP_LINE + INFOPANE_LINE, s, COLOR_YELLOW);
	 if (part_number > 0)
	 {
	    dur = (int)(time(NULL) - rec_partstarttime);
	    hh = dur / 3600;
	    mm = (dur - hh*3600) / 60;
	    ss = dur - hh*3600 - mm*60;
	    sprintf(s, "%d:%02d:%02d", hh, mm, ss);
	    show_text(infofont, 790 - get_text_width(infofont, s), INFOPANE_TOP_LINE + INFOPANE_LINE*2, s, COLOR_YELDIM);
	    sprintf(s, "%s %d", mess[MSG_PART], part_number+1);
	    show_text(infofont, 790 - get_text_width(infofont, s), INFOPANE_TOP_LINE, s, COLOR_CYAN);
	 }

	 show_text(infofont, 790-10*infowidth - get_text_width(infofont, freespace), INFOPANE_TOP_LINE + INFOPANE_LINE*3, freespace,
			lowspace_now?COLOR_WARNRED:COLOR_CYAN);
	 sprintf(s, "%s: %s", mess[MSG_CAPACITY], totalspace);
	 show_text(infofont, 240, INFOPANE_TOP_LINE + INFOPANE_LINE, s, COLOR_WHITE);

	 sprintf(s, "%ld %s", RECfile_MBsize + prev_files_size, (config.menu_language == MENULANG_FR)?"Mo":"MB");
	 show_text(infofont, 790-10*infowidth - get_text_width(infofont, s), INFOPANE_TOP_LINE + INFOPANE_LINE, s, COLOR_WHITE);

	 if (autostop != 0)
	 {
	    min_left = (autostop_time - time(NULL) + 59) / 60;
	    if ((min_left < 0) || (min_left > MAX_AUTOSTOP)) min_left = 0;
	    if (autostop == -1) strcpy(s, "+EPG"); else sprintf(s, "+%dmin", min_left);
	    if (config.poweroff) strcat(s, " !");
	    show_text(infofont, 790-10*infowidth - get_text_width(infofont, s), INFOPANE_TOP_LINE + INFOPANE_LINE*0, s, config.poweroff?COLOR_WARNRED:COLOR_LIGHTRED);
	 }

	 if (part_number > 0)
	 {
	    sprintf(s, "%ld %s", RECfile_MBsize, (config.menu_language == MENULANG_FR)?"Mo":"MB");
	    show_text(infofont, 790-10*infowidth - get_text_width(infofont, s), INFOPANE_TOP_LINE + INFOPANE_LINE*2, s, COLOR_HALFGRAY);
	 }

	 switch(fs.f_type)
	 {
	    case FS_EXT1:
		 strcpy(s2, "EXT1");
		 break;
	    case FS_EXT2_OLD:
		 strcpy(s2, "EXT2 OLD");
		 break;
	    case FS_EXT234:
		 strcpy(s2, "EXT2/3/4");
		 break;
	    case FS_HPFS:
		 strcpy(s2, "HPFS");
		 break;
	    case FS_ISOFS:
		 strcpy(s2, "ISOFS");
		 break;
	    case FS_FAT:
		 strcpy(s2, "FAT");
		 break;
	    case FS_RFS:
		 strcpy(s2, "RFS");
		 break;
	    case FS_CIFS:
		 strcpy(s2, "CIFS");
		 break;
	    case FS_NFS:
		 strcpy(s2, "NFS");
		 break;
	    case FS_NTFS:
		 strcpy(s2, "NTFS");
		 break;
	    case FS_SMB:
		 strcpy(s2, "SMB");
		 break;
	    case FS_UDF:
		 strcpy(s2, "UDF");
		 break;
	    case FS_USBDEV:
		 strcpy(s2, "USB DEV");
		 break;
	    case FS_EFS:
		 strcpy(s2, "EFS");
		 break;
	    default:
		 sprintf(s2, "%X", fs.f_type);
		 break;
	 }

	 strcpy(s, targets[config.current_target].product); if (s[0]==0) strcpy(s, targets[config.current_target].name);
	 while (get_text_width(SHADOW20, s) > 150) s[strlen(s)-2]=0;
	 show_text(SHADOW26, 242, INFOPANE_TOP_LINE-5, s, COLOR_LIGHTGRAY);
	 show_text(SHADOW20, 252 + get_text_width(SHADOW26, s), INFOPANE_TOP_LINE-1, s2, COLOR_GRAY);
	 show_text(infofont, 240, INFOPANE_TOP_LINE + INFOPANE_LINE*3, mess[MSG_FREE_SPACE], COLOR_CYAN);

	 if (bitrate > 0)
	 {
	    if (bitrate < TOO_BIG_BITRATE) sprintf(s, "%s: %u %s/s", mess[MSG_BITRATE], bitrate, (config.menu_language == MENULANG_FR)?"Ko":"KB");
	    else strcpy(s, mess[MSG_UNKBITRATE]);
	    show_text(infofont, 240, INFOPANE_TOP_LINE + INFOPANE_LINE*2, s, COLOR_CYAN);

	    if (free_minutes > 180) sprintf(s, "%ld h", free_minutes/60); else sprintf(s, "%ld min", free_minutes);
	    show_text(infofont, 790 - get_text_width(infofont, s), INFOPANE_TOP_LINE + INFOPANE_LINE*3, s,
				(free_minutes < RED_MINUTES_FREE)?COLOR_WARNRED:COLOR_CYAN);
	 }

	 if (divide_now) show_text(SHADOW20, 241, WARNING_LINE, mess[MSG_DIVIDED], COLOR_CYAN);
	 else if (lowspace_now) show_text(SHADOW20, 241, WARNING_LINE, mess[MSG_LOWSPACE], COLOR_WARNRED);
	 else if (EPG_namebuffer[0])
	 {
	    unicode2utf8(EPG_namebuffer, s, sizeof(s), 0);

	    char sh = 0;
	    while (get_text_width(SHADOW20, s) > 780-241) { s[strlen(s)-1] = 0; sh = 1; }
	    if (sh) strcat(s, "...");
	    show_text(SHADOW20, 241, WARNING_LINE, s, COLOR_GRAY);
	 }

	 SDL_Flip(screen);
      }
   }
}

int infopane_show(int effects)
{
   int x, y, w, h, xpos, ypos, yslide, step;
   unsigned char R, G, B, A = 255;
   unsigned char *scr, *ico_back, *ico;
   SDL_Surface *icon;
   char s[MAX_PATH];

   if (screen)
   {
      sprintf(s, "%s%s/%s.png", config.path, OSD_SUBDIR, ICON_INFOPANE);

      icon = IMG_Load(s);
      if (icon)
      {
	 h = icon->h;
	 w = icon->w;
	 mix_dest_icon(icon);

	 SDL_LockSurface(icon);
	 ico = (unsigned char *)icon->pixels;
	 ico_back = ico;
	 xpos = (screen->w - w) >> 1;

	 step = 16;
	 for (yslide = (effects)?step:(h-1); yslide < h; yslide+=step)
	 {
	    if (yslide > h - 4) step = 1;
	    else
	    if (yslide > h - 8) step = 2;
	    else
	    if (yslide > h - 16) step = 4;

	    ypos = (screen->h - yslide -1);
	    ico = ico_back;
	    SDL_LockSurface(screen);
	    scr = screen->pixels + ypos * screen->pitch + xpos * (SCREEN_BPP/8);
	    for (y = 0; y <= yslide; y++)
            {
                for (x = 0; x < w; x++)
		{
		    R = *(ico++);
		    G = *(ico++);
		    B = *(ico++);
		    A = *(ico++);
		    
		    *(scr++) = B;
		    *(scr++) = G;
		    *(scr++) = R;
		    *(scr++) = A * config.alpha/100 * config.OSD/100;
		}
                scr += screen->pitch - w * (SCREEN_BPP/8);
            }
	    SDL_Flip(screen);
	    SDL_UnlockSurface(screen);
	 }
	    
	 SDL_UnlockSurface(icon);
	 SDL_FreeSurface(icon); icon = NULL;
	 infopane_update = 1;

         return 1;
      }
   }
   return 0;
}

int infopane_hide()
{
   int x, y, w, h, xpos, ypos, dim, offset;
   unsigned char *scr;
   SDL_Surface *icon;
   unsigned char A;
   char s[MAX_PATH];

   infopane_update = 0;

   if (screen)
   {
      sprintf(s, "%s%s/%s.png", config.path, OSD_SUBDIR, ICON_INFOPANE);
   
      icon = IMG_Load(s);
      if (icon)
      {
	 h = icon->h;
	 w = icon->w;
	 SDL_FreeSurface(icon); icon = NULL;

	 xpos = (screen->w - w) >> 1;
	 ypos = screen->h - h;
	 offset = ypos * screen->pitch + xpos * (SCREEN_BPP/8) + 3;

	 if (config.OSD_effects) for (dim = 0; dim < (256/DIM_SPEED); dim++)
	 {
            SDL_LockSurface(screen);
	    scr = screen->pixels + offset;
	    for (y = 0; y < h; y++)
            {
                for (x = 0; x < w; x++)
		{
		    A = *(scr);
		    if (A > DIM_SPEED) A-=DIM_SPEED; else A=0;
		    *(scr) = A;
		    scr+=4;
		}
                scr += screen->pitch - w * (SCREEN_BPP/8);
            }
	    SDL_UnlockSurface(screen);
	    SDL_Flip(screen);
	 }

	 SDL_LockSurface(screen);
	 scr = screen->pixels + offset - 3;
	 for (y = 0; y < h; y++)
         {
             for (x = 0; x < w * (SCREEN_BPP/8); x++) *(scr++) = 0;
             scr += screen->pitch - w * (SCREEN_BPP/8);
         }
	 SDL_UnlockSurface(screen);
	 SDL_Flip(screen);

         return 1;
      }
   }
   return 0;
}

void question_update()
{
   char s[256];

   if (screen)
   {
      config.menu_language = get_menu_language();
      if (config.menu_language != last_menulang) load_messages();

      draw_bitmap(question_but1x, question_but1y, (question_answer==1)?BUTTON_FOCUS:BUTTON_NONSEL, 0);
      draw_bitmap(question_but2x, question_but1y, (question_answer==1)?BUTTON_NONSEL:BUTTON_FOCUS, 0);

      strcpy(s, mess[MSG_OK]);
      show_text(SHADOW20, question_but1x + 54 - get_text_width(SHADOW20, s)/2, question_but1y + 8, s, COLOR_WHITE);
      strcpy(s, mess[MSG_CANCEL]);
      show_text(SHADOW20, question_but2x + 54 - get_text_width(SHADOW20, s)/2, question_but1y + 8, s, COLOR_WHITE);

      SDL_Flip(screen);
   }
}

int show_QUESTION(int msg)
{
   SDL_Rect dstrect;
   SDL_Surface *icon;
   char s[256];

   if (!config.OSD) return KEY_NOTHING;
   if (mode_teletext > 0) return KEY_NOTHING;

   esaving_panel_on();

   init_video_mode();
   if (screen)
   {
      config.menu_language = get_menu_language();
      if (config.menu_language != last_menulang) load_messages();

      sprintf(s, "%s%s/%s.png", config.path, OSD_SUBDIR, ICON_WARNBOX_EXT);
      icon = IMG_Load(s);
      if (icon)
      {
	 dstrect.x = (SCREEN_W - icon->w) >> 1;
	 dstrect.y = (SCREEN_H - icon->h) >> 1;
	 dstrect.w = 477;
	 dstrect.h = 152;
	 SDL_BlitSurface(icon, NULL, screen, &dstrect);

	 question_but1x = dstrect.x + 116;
	 question_but2x = dstrect.x + 252;
	 question_but1y = dstrect.y + 107;

	 SDL_FreeSurface(icon); icon = NULL;
         show_text(SHADOW18, dstrect.x + 85, dstrect.y + 41, mess[msg], COLOR_ALMOSTWHITE);
	 question_answer = -1;
         question_update();
         question = 1;
      }
      return KEY_NOTHING;
   }
   else close_video_mode();
   return KEY_DISP_CABLE;
}

void clear_warninfo()
{
   unsigned char *scr;
   int y, offset;

   if (screen)
   {
      offset = (WARNINFO_Y-3) * screen->pitch + WARNINFO_X * SCREEN_BPP/8;

      scr = screen->pixels + offset;
      SDL_LockSurface(screen);
      for (y = 0; y < FI_BOLD32_HEIGHT+9; y++)
      {
	  memset(scr, 0, (SCREEN_W - WARNINFO_X) * SCREEN_BPP/8);
	  scr += screen->pitch;
      }
      SDL_UnlockSurface(screen);
      SDL_Flip(screen);
   }
}

void show_WARNINFO(int msg)
{
   int x, y, offset, step, pix;
   unsigned char *scr;
   unsigned char alpha;

   if (!config.warnings) return;
   if (!config.OSD) return;
   if (mode_teletext > 0) return;

   while (warninfo_busy) SDL_Delay(100);
   warninfo_busy = 1;

   esaving_panel_on();

   init_video_mode();
   if (screen)
   {
      config.menu_language = get_menu_language();
      if (config.menu_language != last_menulang) load_messages();

      alpha = WARNINFO_BAR_ALPHA * config.OSD/100 * config.alpha/100;

      SDL_LockSurface(screen);
      offset = (WARNINFO_Y-3) * screen->pitch + WARNINFO_X * SCREEN_BPP/8;
      int from_step = (config.OSD_effects)?9:0;
      for (step = from_step; step >= 0; step--)
      {
	  scr = screen->pixels + offset + step*(SCREEN_W - WARNINFO_X) * SCREEN_BPP/80;
	  for (y = 0; y < FI_BOLD32_HEIGHT+9; y++)
	  {
	      pix = (10-step)*(SCREEN_W - WARNINFO_X) / 10;
	      for (x = 0; x < pix; x++)
	      {
		  *(scr++) = WARNINFO_BAR_COLOR;
		  *(scr++) = WARNINFO_BAR_COLOR;
		  *(scr++) = WARNINFO_BAR_COLOR;
		  *(scr++) = alpha;
	      }
	      scr += screen->pitch - pix * SCREEN_BPP/8;
	  }
	  SDL_Flip(screen);
	  SDL_Delay(10);
      }
      SDL_UnlockSurface(screen);

      show_bold_text(WARNINFO_X+15, WARNINFO_Y, mess[msg]);
      SDL_Flip(screen);
      SDL_Delay(WARNINFO_DELAY);
      SDL_LockSurface(screen);
      if (config.OSD_effects)
      {
	 for (step=0; step<10; step++)
	 {
	     scr = screen->pixels + offset;
	     for (y = 0; y < FI_BOLD32_HEIGHT+9; y++)
	     {
		 memset(scr, 0, (SCREEN_W - WARNINFO_X) * SCREEN_BPP/8);
		 scr += screen->pitch;
	     }
	     scr = screen->pixels + offset + step*(SCREEN_W - WARNINFO_X) * SCREEN_BPP/80;
	     for (y = 0; y < FI_BOLD32_HEIGHT+9; y++)
	     {
		 pix = (10-step)*(SCREEN_W - WARNINFO_X) / 10;
		 for (x = 0; x < pix; x++)
		 {
		     *(scr++) = WARNINFO_BAR_COLOR;
		     *(scr++) = WARNINFO_BAR_COLOR;
		     *(scr++) = WARNINFO_BAR_COLOR;
		     *(scr++) = alpha;
		 }
		 scr += screen->pitch - pix * SCREEN_BPP/8;
	     }
	     SDL_Flip(screen);
	     SDL_Delay(8);
	 }
      }
      SDL_UnlockSurface(screen);
      clear_warninfo();
   }
   warninfo_busy = 0;
}

int show_WARNING(int msg, int close_screen)
{
   SDL_Rect dstrect;
   SDL_Surface *icon;
   char s[MAX_PATH];

   if (!config.warnings) return KEY_NOTHING;
   if (!config.OSD) return KEY_NOTHING;
   if (mode_teletext > 0) return KEY_NOTHING;

   esaving_panel_on();

   init_video_mode();
   if (screen)
   {
      config.menu_language = get_menu_language();
      if (config.menu_language != last_menulang) load_messages();
      sprintf(s, "%s%s/%s.png", config.path, OSD_SUBDIR, ICON_WARNBOX);
      icon = IMG_Load(s);
      if (icon)
      {
	 dstrect.x = ((SCREEN_W - icon->w) >> 1) + 1;
	 dstrect.y = (SCREEN_H - icon->h) >> 1;
	 dstrect.w = icon->w;
	 dstrect.h = icon->h;
	 SDL_BlitSurface(icon, NULL, screen, &dstrect);
	 SDL_FreeSurface(icon); icon = NULL;
      }

      char additional_text = ((rec_history > 0) && (msg == MSG_REC_DELETED))?1:0;
      show_text(SHADOW18, dstrect.x + 89, dstrect.y + 42 - 10*additional_text, mess[msg], COLOR_ALMOSTWHITE);

      if (additional_text)
      {
	 sprintf(s, "(%d %s)", rec_history, (rec_history==1)?mess[MSG_FILE]:mess[MSG_FILES]);
	 show_text(SHADOW18, dstrect.x + 89, dstrect.y + 52, s, COLOR_HALFGRAY);
      }
      SDL_Flip(screen);
      SDL_Delay(WARNING_DELAY);

      if (close_screen) close_video_mode();
      return KEY_NOTHING;
   }
   else if (close_screen) close_video_mode();
   return KEY_DISP_CABLE;	
}

void clear_WARNING()
{
   SDL_Surface *icon;
   unsigned char *scr;
   int y;
   char s[MAX_PATH];

   if (screen)
   {
      sprintf(s, "%s%s/%s.png", config.path, OSD_SUBDIR, ICON_WARNBOX);
      icon = IMG_Load(s);
      if (icon)
      {
         SDL_LockSurface(screen);
         scr = screen->pixels + ((SCREEN_H - icon->h) >> 1) * screen->pitch;
         for (y = 0; y < icon->h; y++)
         {
	     memset(scr, 0, SCREEN_W * SCREEN_BPP/8);
	     scr += screen->pitch;
         }
         SDL_UnlockSurface(screen);
         SDL_Flip(screen);
	 SDL_FreeSurface(icon); icon = NULL;
      }
   }
}

void show_AUTOSTOP()
{
   char s[80];
   int i, ln, min_left;

   if (!config.OSD) return;
   if (!is_recording) return;
   if (config.teletext) if (mode_teletext > 0) return;

   while (warninfo_busy) SDL_Delay(100);
   warninfo_busy = 1;

   init_video_mode();
   if (screen)
   {
      config.menu_language = get_menu_language();
      if (config.menu_language != last_menulang) load_messages();

      min_left = (autostop_time - time(NULL) + 59) / 60;
      if ((min_left < 0) || (min_left > MAX_AUTOSTOP)) min_left = 0;
      if (autostop == 0) sprintf(s, " %s ", mess[MSG_OFF]);
      else if (autostop == -1)
      {
	 sprintf(s, " %s ", mess[MSG_STOP_BY_EPG]);
	 if (config.EPG_postminutes > 0)
	    sprintf(s+strlen(s), "+%dmin ", config.EPG_postminutes);
      }
      else
      {
	 sprintf(s, " +%dmin ", min_left);

	 int cur_year, cur_hour, cur_min;
	 if ((config.timesource == TIMESOURCE_EPG) && EPG_is_initialized)
	 {
	    cur_year = EPG_year;
	    cur_hour = EPG_hour; cur_min = EPG_minute;
	 }
	 else
	 {
	    struct tm *tm;
	    tm = localtime_from_UTC_system(timezones[config.timezone].winter_time, timezones[config.timezone].summer_time);
	    cur_year = tm->tm_year; if (cur_year<1900) cur_year+=1900;
	    cur_hour = tm->tm_hour;
	    cur_min  = tm->tm_min;
	 }
	 if ((cur_year >= MIN_YEAR) && (cur_year <= MAX_YEAR))
	 {
	    int end_hh = cur_hour;
	    int end_mm = cur_min + min_left;
	    if (end_mm >= 60)
	    {
	       int add_hh = end_mm/60;
	       end_mm -= add_hh*60;
	       end_hh += add_hh;
	       if (end_hh >= 24)
	       {
		  int add_dd = end_hh/24;
		  end_hh -= add_dd*24;
	       }
	    } 
	    if ((end_hh >= 0) && (end_hh < 24) && (end_mm >=0) && (end_mm < 60))
		sprintf(s+strlen(s), "(%02d:%02d) ", end_hh, end_mm);
	 }
      }
      ln = strlen(s); if (ln > 0) for (i = 0; i < ln; i++) if (s[i]==' ') s[i] = SPC;
      show_bold_text(WARNINFO_X+15, WARNINFO_Y, s);
      SDL_Flip(screen);
      SDL_Delay(1000);

      clear_warninfo();	
   }
   warninfo_busy = 0;
}

int set_icon(const char *pngname)
{
   int x, y;
   SDL_Surface *icon;
   unsigned char R, G, B, A = 255;
   char s[MAX_PATH];

   if (config.OSD == 0) return 1;
   if (mode_teletext > 0) return 1;

   if (screen)
   {
      sprintf(s, "%s%s/%s.png", config.path, OSD_SUBDIR, pngname);
   
      icon = IMG_Load(s);
      if (icon)
      {
	    SDL_LockSurface(screen);
	    SDL_LockSurface(icon);
	    unsigned char *scr = (unsigned char *)screen->pixels;
	    unsigned char *ico = (unsigned char *)icon->pixels;
	    int ico_bpp = icon->format->BytesPerPixel;
	    int alpha2 = config.alpha * config.OSD;
	    if (pngname == ICON_STOPREC) alpha2 /= 2;

	    scr += ICON_Y * screen->pitch + ICON_X * (SCREEN_BPP/8);
	    for (y = 0; y < icon->h; y++)
            {
                for (x = 0; x < icon->w; x++)
		{
		    R = *(ico++);
		    G = *(ico++);
		    B = *(ico++);
		    if (ico_bpp == 4) A = *(ico++);
		    
		    *(scr++) = B;
		    *(scr++) = G;
		    *(scr++) = R;
		    *(scr++) = A * alpha2/10000;
		}
                scr += screen->pitch - icon->w * (SCREEN_BPP/8);
            }
	    SDL_UnlockSurface(icon);
	    SDL_UnlockSurface(screen);
	    SDL_Flip(screen);

	    SDL_FreeSurface(icon); icon = NULL;

            return 1;
      }
   }
   return 0;
}

int show_NO()
{
   if (!config.warnings) return KEY_NOTHING;
   if (!config.OSD) return KEY_NOTHING;
   if (mode_teletext > 0) return KEY_NOTHING;

   if (screen)
   {
      close_video_mode();

      init_video_mode();
      if (screen)
      {
         if (set_icon(ICON_NO)) SDL_Delay(NO_DELAY);

         close_video_mode();

         init_video_mode();
         if (screen)
         {
            if (recdispmode == RECDISPMODE_HIDDEN) OSD_cls();
	    else set_icon(ICON_REC);
         }
      }

      return KEY_NOTHING;
   }
   else close_video_mode();
   return KEY_DISP_CABLE;
}

int show_REC()
{
   if (!config.OSD) return KEY_NOTHING;
   if (mode_teletext > 0) return KEY_NOTHING;

   init_video_mode();
   if (screen)
   {
      if (!set_icon(ICON_REC))
      {
         close_video_mode();
         return KEY_DISP_VCR;
      }
      return KEY_NOTHING;
   }
   else close_video_mode();

   return KEY_DISP_VCR;
}

void show_SCRIPT()
{
   if (!config.OSD) return;
   if (mode_teletext > 0) return;

   init_video_mode();
   if (screen)
   {
      if (!set_icon(ICON_SCRIPT))
      {
         close_video_mode();
         return;
      }
      return;
   }
   else close_video_mode();
}

int stop_REC(int recording_deleted)
{
   if (!config.OSD) return KEY_NOTHING;
   if (mode_teletext > 0) return KEY_NOTHING;

   if (screen)
   {
      if (!emergency_stop)
      {
         close_video_mode();
      }
      init_video_mode();
      if (screen)
      {
	 set_icon(ICON_STOP);

         if (recording_deleted)
         {
            show_WARNING(MSG_DELETED, 1);
         }
         else
         {
            SDL_Delay(STOP_DELAY);
         }

	 if (!emergency_stop)
            close_video_mode();
      }
      return KEY_NOTHING;
   }
   else
   return KEY_DISP_TV;
}

void show_summary()
{
   char s[256];

   if (config.OSD == 0) return;
   if (mode_teletext > 0) return;

   init_video_mode();
   if (screen)
   {
      config.menu_language = get_menu_language();
      if (config.menu_language != last_menulang) load_messages();

      draw_bitmap((SCREEN_W - 530)/2, SCREEN_H - 220, ICON_TOOLBOX_HEADER, 0);
      change_alpha((SCREEN_W - 530)/2+2, SCREEN_H - 220 + 38, 530-4, 120, INSIDE_ALPHA * config.OSD/100);
      strcpy(s, mess[MSG_SUMMARY]);
      show_text(SHADOW20, (SCREEN_W - get_text_width(SHADOW20, s))/2, SCREEN_H - 211, s, COLOR_WHITE);

      unsigned long rt = (int)(time(NULL) - rec_starttime);
      int hh = rt / 3600;
      int mm = (rt - hh*3600) / 60;
      int ss = (rt - hh*3600 - mm*60);
      sprintf(s, "%s: %d:%02d:%02d", mess[MSG_TOTALRECTIME], hh,mm,ss);
      show_text(SHADOW18, (SCREEN_W - 500)/2, SCREEN_H - 170, s, COLOR_YELLOW);

      sprintf(s, "%s: %lu %s", mess[MSG_TOTALRECSIZE], RECfile_MBsize + prev_files_size, (config.menu_language == MENULANG_FR)?"Mo":"MB");
      show_text(SHADOW18, (SCREEN_W - 500)/2, SCREEN_H - 170 + FI_SHADOW18_HEIGHT, s, COLOR_LIGHTRED);

      sprintf(s, "%s: %d", mess[MSG_NUMOFPARTS], part_number+1);
      show_text(SHADOW18, (SCREEN_W - 500)/2, SCREEN_H - 170 + FI_SHADOW18_HEIGHT*2, s, COLOR_CYAN);

      sprintf(s, "%s", mess[MSG_LASTFILENAME]);
      if (config.DVRnr > 0)
	 sprintf(s+strlen(s), " (%s %s%d%s)", mess[MSG_INSUBDIR1], DVR_SUBDIR_NAME, config.DVRnr, mess[MSG_INSUBDIR2]);
      strcat(s, ":");
      show_text(SHADOW18, (SCREEN_W - 500)/2, SCREEN_H - 170 + FI_SHADOW18_HEIGHT*3, s, COLOR_LIGHTGRAY);
      int tl = strlen(ts_filename);
      int fc, ffc = 0;
      if (tl > 0)
      {
	 for (fc=0; fc < tl; fc++) if ((ts_filename[fc] == '/') || (ts_filename[fc] == '\\')) ffc = fc+1;
	 strcpy(s, ts_filename+ffc);
	 tl = strlen(s); fc = tl;
	 if (tl > 0)
	 {
	    while (get_text_width(SHADOW20, s) > LASTFILENAME_MAX_PIXELS) s[--tl]=0;
	    if (fc != tl) strcat(s, "...");
	    show_text(SHADOW20, (SCREEN_W - 500)/2, SCREEN_H - 170 + FI_SHADOW18_HEIGHT*4, s, COLOR_WHITE);
	 }
      }

      sprintf(s, "- %s", mess[MSG_DELBUTTON]);
      tl = get_text_width(SHADOW18, s);
      draw_bitmap((SCREEN_W - 530)/2 + 520 - tl - 30, SCREEN_H - 50, ICON_RED, 0);
      show_text(SHADOW18, (SCREEN_W - 530)/2 + 512 - tl, SCREEN_H - 53, s, COLOR_ALMOSTWHITE);

      SDL_Flip(screen);
      SDL_Delay(SUMMARY_DELAY);
      if (emergency_stop) OSD_cls(); else close_video_mode();
   }
}

int send_button(int key)
{
   unsigned *pInstance = (unsigned *)(config.pKeyInputBase[2]);
   config.pKeyInputBase = (unsigned *)(*pInstance);
   if (config.pKeyInputBase[6] != (unsigned)config.KeyInputBase_SendKey) return 0;

   asm volatile(
		"MOV	R0,%0\n"
		"MOV 	R1,%1\n"
		"BL	_ZN14SsKeyInputBase7SendKeyEi\n"
		:
		: "r" (pInstance), "r" (key) );

   return 1;
}

void power_off()
{
   show_WARNINFO(MSG_POWEROFF);
   send_button(KEY_BEZEL_POWER);
}

int total_free_space(int target_nr)
{
   struct statvfs vfs;

   if (statvfs(targets[target_nr].path, &vfs) != 0) return 0;

   double d_bavail = vfs.f_bavail;
   double d_bsize  = vfs.f_bsize;
   double d_blocks = vfs.f_blocks;

   double dKBfree = d_bavail * d_bsize / 1024;
   KBfree = (unsigned long)dKBfree;

   double dKBtotal = d_blocks * d_bsize / 1024;
   KBtotal = (unsigned long)dKBtotal;

   return 1;
}

unsigned long get_file_MBsize(char *fname)
{
   struct stat fst;

   if (stat(fname, &fst) == 0)
   {
      return (unsigned long)(fst.st_size / 1024 / 1024);
   }
   return 0L;
}

Uint32 rec_monitoring(Uint32 interval, void *param)
{
   monitoring_now = 1;

   if (!is_recording) { monitoring_now = 0; return 0; }

   if (autostop > 0)
      if (time(NULL) >= autostop_time)
      {
	 autostop = 0;
	 show_WARNINFO(MSG_AUTOSTOP);
	 emergency_stop = 1;
	 button_pressed(-KEY_STOP);

	 emergency_stop = 2;
	 OSD_cls();
	 if (config.poweroff) power_off();
	 return 0;
      }

   if (totalfree_step == 0)
   {
      if (total_free_space(config.current_target))
      {
	 double MBfree  = (double)KBfree  / 1024;
	 double GBfree  = (double)MBfree  / 1024;
	 double MBtotal = (double)KBtotal / 1024;
	 double GBtotal = (double)MBtotal / 1024;

	 if (GBfree > 2) sprintf(freespace, "%0.2f %s", GBfree, (config.menu_language == MENULANG_FR)?"Go":"GB");
	 else if (MBfree > 8) sprintf(freespace, "%0.2f %s", MBfree, (config.menu_language == MENULANG_FR)?"Mo":"MB");
	 else sprintf(freespace, "%lu %s", KBfree, (config.menu_language == MENULANG_FR)?"Ko":"KB");

	 if (GBtotal > 2) sprintf(totalspace, "%0.2f %s", GBtotal, (config.menu_language == MENULANG_FR)?"Go":"GB");
	 else if (MBtotal > 8) sprintf(totalspace, "%0.2f %s", MBtotal, (config.menu_language == MENULANG_FR)?"Mo":"MB");
	 else sprintf(totalspace, "%lu %s", KBtotal, (config.menu_language == MENULANG_FR)?"Ko":"KB");

	 free_level = (int)((double)(FREE_LEVELS-1) * KBfree / KBtotal);

	 if (MBfree <= FREE_SPACE_STOP)
	 {
	    show_WARNINFO(MSG_NOSPACE);
	    emergency_stop = 1;
	    button_pressed(-KEY_STOP);

	    emergency_stop = 2;
	    OSD_cls();
	    if (config.poweroff) power_off();
	    return 0;
	 }

	 if (MBfree < FREE_SPACE_WARNING)
	 {
	    if (!lowspace_warned)
	    {
	       lowspace_now = 1;
	       show_WARNINFO(MSG_LOWSPACE);
	       lowspace_warned = 1;
	    }
	 }
	 else { lowspace_warned = 0; lowspace_now = 0; }
      }

      RECfile_MBsize = get_file_MBsize(ts_filename);

      unsigned long dur = (unsigned long)(time(NULL) - ((part_number > 0)?rec_partstarttime:rec_starttime));
      if (dur > 0) bitrate = (unsigned)(RECfile_MBsize*1024 / dur);

      if ((bitrate > 0) && (KBfree >= FREE_SPACE_STOP*1024))
	 free_minutes = (KBfree - FREE_SPACE_STOP*1024) / bitrate / 60;

      if (current_divide != DIVIDING_OFF)
      {
	 if (current_divide != DIVIDING_BY_EPG)
	 {
	    if (RECfile_MBsize > divides[current_divide].val)
	    {
	       divide_now = 1;
	       button_pressed(-KEY_STOP);
	       button_pressed(-KEY_RECORD);
	    }
	    else if (RECfile_MBsize + DIVIDE_WARN_MARGIN > divides[current_divide].val)
	    {
	       if (!divide_warned)
	       {
		  int panel_state = panel_off;
		  show_WARNINFO(MSG_DIVIDEWARN);
		  divide_warned = 1;
		  if (panel_state) esaving_panel_off();
	       }
	    }
	 }
      }
   }


   if ((epg_step == 0) || (EPG_progname_changed > 0))
   {
      int epg = 0;
      epg = EPG_get_data();

      if ((EPG_progname_changed >= EPG_PROGNAME_CHANGED_LIMIT) && (autostop == -1))
      {
	 EPG_progname_changed = 0;
	 EPG_progname_checks = 0;

	 if (config.EPG_postminutes > 0)
	 {
	    autostop = 3;
	    autostop_time = time(NULL) + config.EPG_postminutes*60;
	    show_WARNINFO(MSG_POSTMINUTES);
	 }
	 else
	 {
	    autostop = 0;
	    show_WARNINFO(MSG_AUTOSTOP);
	    emergency_stop = 1;
	    button_pressed(-KEY_STOP);
	    emergency_stop = 2;
	    OSD_cls();
	    if (config.poweroff) power_off();
	    return 0;
	 }
      }

      if ((EPG_progname_changed >= EPG_PROGNAME_CHANGED_LIMIT) && (current_divide == DIVIDING_BY_EPG))
      {
	 EPG_progname_changed = 0;
	 EPG_progname_checks = 0;
	 
	 divide_now = 1;
	 button_pressed(-KEY_STOP);
	 SDL_Delay(200);
	 button_pressed(-KEY_RECORD);
	 SDL_Delay(200);
      }
      else if (epg_step == 0)
      {
	 if (epg_probes == next_probe)
	 {
	    if (next_probe == 0)
	    {
	       next_probe = 1;
	       strcpy(recpartname_prev, epg?recpartname:"");
	       strcpy(recpartname_best, epg?recpartname:"");
	    }
	    else
	    {
	       if (next_probe > 1) if (recpartname_prev[0]) strcpy(recpartname_best, recpartname_prev);
	       if (epg) strcpy(recpartname_prev, recpartname);
	       next_probe *=2;
	    }
	 }
	 epg_probes++;
      }
   }

   if (recwarnings_step == 0)
      if ((config.OSD) && (config.warnings) && (config.recwarnings) && panel_off)
      {
	 if ((recwarned < 3) || ((recwarned & 1) == 0))
	 {
	    if (!video_muted) esaving_togglevideo();
	    esaving_panel_on();
	    show_WARNINFO(MSG_RECINPROGRESS);
	    esaving_panel_off();
	    esaving_togglevideo();
	 }
	 recwarned++;
      }

   totalfree_step++;
   if (totalfree_step >= TOTALFREE_PERIOD) totalfree_step = 0;

   epg_step++;
   if (epg_step >= EPG_PERIOD) epg_step = 0;

   recwarnings_step++;
   if (recwarnings_step >= RECWARNINGS_PERIOD) recwarnings_step = 0;

   if (infopane_update) infopane_update_thread();

   if (divide_now)
   {
      int panel_state = panel_off;
      show_WARNINFO(MSG_DIVIDED);
      divide_now = 0;
      if (panel_state) esaving_panel_off();
   }

   if (is_recording) return MONITORING_INTERVAL; else { monitoring_now = 0; return 0; } 
}

int is_grayed_out(int mess_ID)
{
   switch (mess_ID)
   {
      case SI_TARGET1:
	   break;
      case SI_TARGET2:
	   if (config.target1.id == 0) return 1;
	   break;
      case SI_CREATEINTDIRS:
	   if (!config.useintmem) return 1;
      case SI_USEINTMEM:
	   if ((!targets[config.target1.id].usesintmem) && (!targets[config.target2.id].usesintmem)) return 1;
	   break;
      case SI_OSD_EFFECTS:
      case SI_WARNINGS:
      case SI_INFOPANE:
      case SI_INFOPANEFONT:
	   if (config.OSD == 0) return 1;
	   break;
      case SI_RECWARNINGS:
	   if ((config.warnings == 0) || (config.OSD == 0)) return 1;
	   break;
      case SI_PRE_SCRIPTS:
	   if (config.target1.id == 0) return 1;
	   break;
      default:
	   break;
   }
   return 0;
}

void item_ON_OFF(int px, int py, int config_item, int color)
{
   show_text(SHADOW20, px, py, config_item?mess[MSG_ON]:mess[MSG_OFF], color);
}

void item_YES_NO(int px, int py, int config_item, int color)
{
   show_text(SHADOW20, px, py, config_item?mess[MSG_YES]:mess[MSG_NO], color);
}

void refresh_setup_window()
{
   int i, w, px, py;
   char s[256];

   config.menu_language = get_menu_language();
   if (config.menu_language != last_menulang) load_messages();

   draw_bitmap((SCREEN_W - 540)/2, 45, ICON_TOOLBOX_HEADER, 212);
   change_alpha((SCREEN_W - 540)/2+2, 45+38, 530-4, 369, INSIDE_ALPHA * config.OSD/100);
   draw_bitmap((SCREEN_W - 540)/2, SCREEN_H - 45-41, ICON_TOOLBOX_FOOTER, 0);
   i = get_text_width(SHADOW20, mess[MSG_EXIT]);
   show_text(SHADOW20, (SCREEN_W - 540)/2 + 510-i, SCREEN_H - 77, mess[MSG_EXIT], COLOR_ALMOSTWHITE);
   i+=30; draw_bitmap((SCREEN_W - 540)/2 + 510-i, SCREEN_H - 74, ICON_EXIT, -1);
   i+=20;
   i+= get_text_width(SHADOW20, mess[MSG_SAVE]);
   show_text(SHADOW20, (SCREEN_W - 540)/2 + 510-i, SCREEN_H - 77, mess[MSG_SAVE], COLOR_ALMOSTWHITE);
   i+=26; draw_bitmap((SCREEN_W - 540)/2 + 510-i, SCREEN_H - 76, ICON_ENTER, -1);
   i+=20;
   i+= get_text_width(SHADOW20, mess[MSG_ADJUST]);
   show_text(SHADOW20, (SCREEN_W - 540)/2 + 510-i, SCREEN_H - 77, mess[MSG_ADJUST], COLOR_ALMOSTWHITE);
   i+=22; draw_bitmap((SCREEN_W - 540)/2 + 510-i, SCREEN_H - 80, ICON_LEFTRIGHT, -1);
   i+=20;
   i+= get_text_width(SHADOW20, mess[MSG_MOVE]);
   show_text(SHADOW20, (SCREEN_W - 540)/2 + 510-i, SCREEN_H - 77, mess[MSG_MOVE], COLOR_ALMOSTWHITE);
   i+=20; draw_bitmap((SCREEN_W - 540)/2 + 510-i, SCREEN_H - 80, ICON_UPDOWN, -1);
   strcpy(s, mess[MSG_RECSET]);

   w = get_text_width(SHADOW20, s);
   show_text(SHADOW20, (SCREEN_W - w)/2, 55, s, COLOR_ALMOSTWHITE);
   draw_bitmap((SCREEN_W - 510)/2,        96 + setup_cursor*39, ICON_TOOLBOX_FRAME, -1);
   draw_bitmap((SCREEN_W - 500)/2 + 290, 106 + setup_cursor*39, ICON_LEFT, -1);
   draw_bitmap((SCREEN_W - 500)/2 + 468, 106 + setup_cursor*39, ICON_RIGHT, -1);
   if (setup_offset > 0) draw_bitmap((SCREEN_W - 500)/2, 60, ICON_UP, -1);
   if (setup_offset + MAX_SETUP_ITEMS <= SETUP_ITEMS-1) draw_bitmap((SCREEN_W - 500)/2, SCREEN_H - 73, ICON_DOWN, -1);
   for (i=0; i < MAX_SETUP_ITEMS; i++)
   {
       int item = i + setup_offset;
       int color = (is_grayed_out(item + SI_START))?COLOR_GRAY:COLOR_ALMOSTWHITE;

       show_text(SHADOW20, (SCREEN_W - 480)/2, 106 + i*39, mess[item + SI_START], color);
       if (i != setup_cursor) show_text(SHADOW20, (SCREEN_W - 500)/2 + 300, 106 + i*39, ":", color);
       px = (SCREEN_W - 500)/2 + 320;
       py = 106 + i*39;
       switch(item + SI_START)
       {
	  case SI_LOCK:
		item_ON_OFF(px, py, config.record_key_lock, color);
		break;
	  case SI_TARGET1:
		strcpy(s, targets[config.target1.id].product); if (s[0]==0) strcpy(s, targets[config.target1.id].name);
		w=0; while (get_text_width(SHADOW20, s) > 135) { w=1; s[strlen(s)-2]=0;	}
		if (w) strcat(s, "...");
		show_text(SHADOW20, px, py, s, (config.target1.not_allowed)?COLOR_WARNRED:color);
		break;
	  case SI_TARGET2:
		strcpy(s, targets[config.target2.id].product); if (s[0]==0) strcpy(s, targets[config.target2.id].name);
		w=0; while (get_text_width(SHADOW20, s) > 135) { w=1; s[strlen(s)-2]=0;	}
		if (w) strcat(s, "...");
		show_text(SHADOW20, px, py, s, (config.target2.not_allowed)?COLOR_WARNRED:color);
		break;
	  case SI_PRE_SCRIPTS:
		item_ON_OFF(px, py, config.pre_scripts, color);
		break;
	  case SI_DIVIDE:
		show_text(SHADOW20, px, py, mess[divides[config.divide].option], color);
		break;
	  case SI_EPG_POSTMINUTES:
		if (config.EPG_postminutes == 0) strcpy(s, mess[MSG_OFF]); else sprintf(s, "%d min", config.EPG_postminutes);
		show_text(SHADOW20, px, py, s, color);
		break;
	  case SI_POWEROFF:
		item_ON_OFF(px, py, config.poweroff, color);
		break;
	  case SI_DEL_EMPTY:
		item_ON_OFF(px, py, config.del_empty_ts, color);
		break;
	  case SI_DEL_MDB:
		item_ON_OFF(px, py, config.del_mdb, color);
		break;
	  case SI_TESTWRITE:
		item_ON_OFF(px, py, config.testwrite, color);
		break;
	  case SI_INFOPANE:
		show_text(SHADOW20, px, py, mess[manualauto[config.infopane]], color);
		break;
	  case SI_INFOPANEFONT:
		show_text(SHADOW20, px, py, infopanefonts[config.infopanefont].name, color);
		break;
	  case SI_SUMMARY:
		item_ON_OFF(px, py, config.summary, color);
		break;
	  case SI_TIMESOURCE:
		show_text(SHADOW20, px, py, mess[timesources[config.timesource]], color);
		break;
	  case SI_TIMEZONE:
		show_text(SHADOW20, px, py, timezones[config.timezone].name, color);
		break;
	  case SI_OSD:
		if (config.OSD==0) strcpy(s, mess[MSG_OFF]); else sprintf(s, "%d%%", config.OSD);
		show_text(SHADOW20, px, py, s, color);
		break;
	  case SI_OSD_EFFECTS:
		item_ON_OFF(px, py, config.OSD_effects, color);
		break;
	  case SI_WARNINGS:
		item_ON_OFF(px, py, config.warnings, color);
		break;
	  case SI_RECWARNINGS:
		item_ON_OFF(px, py, config.recwarnings, color);
		break;
	  case SI_SOURCE:
		show_text(SHADOW20, px, py, mess[sources[config.source].option], color);
		break;
	  case SI_FAVCH:
		show_text(SHADOW20, px, py, mess[favchs[config.favch].option], color);
		break;
	  case SI_FREEZE:
		show_text(SHADOW20, px, py, mess[freezes[config.freeze]], color);
		break;
	  case SI_MUTE:
		item_ON_OFF(px, py, config.mute, color);
		break;
	  case SI_TELETEXT:
		item_YES_NO(px, py, config.teletext, color);
		break;
	  case SI_USEINTMEM:
		item_YES_NO(px, py, config.useintmem, color);
		break;
	  case SI_CREATEINTDIRS:
		item_YES_NO(px, py, config.createintdirs, color);
		break;
	  case SI_TRANSCODE:
		show_text(SHADOW20, px, py, mess[EPG_trans[config.EPG_transcode]], color);
		break;
	  default:
		break;
       }
   }
   SDL_Flip(screen);
}

int setup()
{
   if (setup_menu) return KEY_EXIT;

   setup_menu = 1;
   selected_setup_item = 0;
   setup_cursor = 0;
   setup_offset = 0;
   parse_usb_log();
   refresh_config_targets();
   init_video_mode();
   refresh_setup_window();

   return KEY_EXIT;
}

void save_config()
{
   FILE *cfg;
   char filename[MAX_PATH];
   int err = 0;

   sprintf(filename, "%s%s", config.path, CONFIG_FNAME);

   cfg = fopen(filename, "wb");
   if (cfg)
   {
      if (fwrite(&config, sizeof(config), 1, cfg) != 1) err = 1;
      fclose(cfg);
   }

   if (err) remove(filename);
}

void close_setup_window(int save)
{
   close_video_mode();
   setup_menu = 0;
   no_dir1 = 0;

   store_config_target1();
   store_config_target2();

   if (save) save_config();
}

int setup_left()
{
   switch(selected_setup_item + SI_START)
   {
	case SI_LOCK:
		if (config.record_key_lock) config.record_key_lock = 0; else config.record_key_lock = 1;
		break;
	case SI_TARGET1:
		if (config.target1.id == 0) config.target1.id = max_targets-1; else config.target1.id--;
		store_config_target1();
		break;
	case SI_TARGET2:
		if (config.target2.id == 0) config.target2.id = max_targets-1; else config.target2.id--;
		store_config_target2();
		break;
	case SI_PRE_SCRIPTS:
		if (config.pre_scripts) config.pre_scripts = 0; else config.pre_scripts = 1;
		break;
	case SI_DIVIDE:
		if (config.divide == 0) config.divide = MAX_DIVIDES-1; else config.divide--;
		break;
	case SI_EPG_POSTMINUTES:
		if (config.EPG_postminutes > 0) config.EPG_postminutes--; else config.EPG_postminutes = MAX_EPG_POSTMINUTES;
		break;
	case SI_POWEROFF:
		if (config.poweroff) config.poweroff = 0; else config.poweroff = 1;
		break;
	case SI_DEL_EMPTY:
		if (config.del_empty_ts) config.del_empty_ts = 0; else config.del_empty_ts = 1;
		break;
	case SI_DEL_MDB:
		if (config.del_mdb) config.del_mdb = 0; else config.del_mdb = 1;
		break;
	case SI_TESTWRITE:
		if (config.testwrite) config.testwrite = 0; else config.testwrite = 1;
		break;
	case SI_INFOPANE:
		if (config.infopane == 0) config.infopane = MAX_INFOPANE-1; else config.infopane--;
		break;
	case SI_INFOPANEFONT:
		if (config.infopanefont == 0) config.infopanefont = MAX_INFOPANEFONT-1; else config.infopanefont--;
		break;
	case SI_SUMMARY:
		if (config.summary) config.summary = 0; else config.summary = 1;
		break;
	case SI_TIMESOURCE:
		if (config.timesource == 0) config.timesource = MAX_TIMESOURCES-1; else config.timesource--;
		break;
	case SI_TIMEZONE:
		if (config.timezone == 0) config.timezone = MAX_TIMEZONES-1; else config.timezone--;
		break;
	case SI_OSD:
		if (config.OSD == 0) config.OSD = 100; else config.OSD -= 10;
		break;
	case SI_OSD_EFFECTS:
		if (config.OSD_effects) config.OSD_effects = 0; else config.OSD_effects = 1;
		break;
	case SI_WARNINGS:
		if (config.warnings) config.warnings = 0; else config.warnings = 1;
		break;
	case SI_RECWARNINGS:
		if (config.recwarnings) config.recwarnings = 0; else config.recwarnings = 1;
		break;
	case SI_SOURCE:
		if (config.source == 0) config.source = MAX_SOURCES-1; else config.source--;
		break;
	case SI_FAVCH:
		if (config.favch == 0) config.favch = MAX_FAVCHS-1; else config.favch--;
		break;
	case SI_FREEZE:
		if (config.freeze == 0) config.freeze = MAX_FREEZES-1; else config.freeze--;
		break;
	case SI_MUTE:
		if (config.mute) config.mute  = 0; else config.mute = 1;
		break;
	case SI_TELETEXT:
		if (config.teletext) config.teletext = 0; else config.teletext = 1;
		break;
	case SI_USEINTMEM:
		if (config.useintmem) config.useintmem = 0; else config.useintmem = 1;
		break;
	case SI_CREATEINTDIRS:
		if (config.createintdirs) config.createintdirs = 0; else config.createintdirs = 1;
		break;
	case SI_TRANSCODE:
		if (config.EPG_transcode == 0) config.EPG_transcode = MAX_EPG_TRANS-1; else config.EPG_transcode--;
		break;
	default:
		break;
   }
   refresh_setup_window();
   return KEY_NOTHING;
}

int setup_right()
{
   switch(selected_setup_item + SI_START)
   {
	case SI_LOCK:
		if (config.record_key_lock) config.record_key_lock = 0; else config.record_key_lock = 1;
		break;
	case SI_TARGET1:
		if (config.target1.id >= max_targets-1) config.target1.id = 0; else config.target1.id++;
		store_config_target1();
		break;
	case SI_TARGET2:
		if (config.target2.id >= max_targets-1) config.target2.id = 0; else config.target2.id++;
		store_config_target2();
		break;
	case SI_PRE_SCRIPTS:
		if (config.pre_scripts) config.pre_scripts = 0; else config.pre_scripts = 1;
		break;
	case SI_DIVIDE:
		if (config.divide >= MAX_DIVIDES-1) config.divide = 0; else config.divide++;
		break;
	case SI_EPG_POSTMINUTES:
		if (config.EPG_postminutes < MAX_EPG_POSTMINUTES) config.EPG_postminutes++; else config.EPG_postminutes = 0;
		break;
	case SI_POWEROFF:
		if (config.poweroff) config.poweroff = 0; else config.poweroff = 1;
		break;
	case SI_DEL_EMPTY:
		if (config.del_empty_ts) config.del_empty_ts = 0; else config.del_empty_ts = 1;
		break;
	case SI_DEL_MDB:
		if (config.del_mdb) config.del_mdb = 0; else config.del_mdb = 1;
		break;
	case SI_TESTWRITE:
		if (config.testwrite) config.testwrite = 0; else config.testwrite = 1;
		break;
	case SI_INFOPANE:
		if (config.infopane >= MAX_INFOPANE-1) config.infopane = 0; else config.infopane++;
		break;
	case SI_INFOPANEFONT:
		if (config.infopanefont >= MAX_INFOPANEFONT-1) config.infopanefont = 0; else config.infopanefont++;
		break;
	case SI_SUMMARY:
		if (config.summary) config.summary = 0; else config.summary = 1;
		break;
        case SI_TIMESOURCE:
		if (config.timesource >= MAX_TIMESOURCES-1) config.timesource = 0; else config.timesource++;
		break;
        case SI_TIMEZONE:
		if (config.timezone >= MAX_TIMEZONES-1) config.timezone = 0; else config.timezone++;
		break;
	case SI_OSD:
		if (config.OSD >= 100) config.OSD = 0; else config.OSD += 10;
		break;
	case SI_OSD_EFFECTS:
		if (config.OSD_effects) config.OSD_effects = 0; else config.OSD_effects = 1;
		break;
	case SI_WARNINGS:
		if (config.warnings) config.warnings = 0; else config.warnings = 1;
		break;
	case SI_RECWARNINGS:
		if (config.recwarnings) config.recwarnings = 0; else config.recwarnings = 1;
		break;
	case SI_SOURCE:
		if (config.source >= MAX_SOURCES-1) config.source = 0; else config.source++;
		break;
	case SI_FAVCH:
		if (config.favch >= MAX_FAVCHS-1) config.favch = 0; else config.favch++;
		break;
	case SI_FREEZE:
		if (config.freeze >= MAX_FREEZES-1) config.freeze = 0; else config.freeze++;
		break;
	case SI_MUTE:
		if (config.mute) config.mute  = 0; else config.mute = 1;
		break;
	case SI_TELETEXT:
		if (config.teletext) config.teletext = 0; else config.teletext = 1;
		break;
	case SI_USEINTMEM:
		if (config.useintmem) config.useintmem = 0; else config.useintmem = 1;
		break;
	case SI_CREATEINTDIRS:
		if (config.createintdirs) config.createintdirs = 0; else config.createintdirs = 1;
		break;
	case SI_TRANSCODE:
		if (config.EPG_transcode >= MAX_EPG_TRANS-1) config.EPG_transcode = 0; else config.EPG_transcode++;
		break;
	default:
		break;
   }
   refresh_setup_window();
   return KEY_NOTHING;
}

int setup_up()
{
   if (selected_setup_item > 0)
   {
      do
      {
         selected_setup_item--;
         if (setup_cursor > 0) setup_cursor--;
         else if (setup_offset > 0) setup_offset--;
      } while (is_grayed_out(selected_setup_item + SI_START));
      if (selected_setup_item < 0)
	 { selected_setup_item = prev_selected_setup_item; setup_cursor = prev_setup_cursor; setup_offset = prev_setup_offset; }
      refresh_setup_window();
      prev_selected_setup_item = selected_setup_item; prev_setup_cursor = setup_cursor; prev_setup_offset = setup_offset;
   }
   else
   {
      selected_setup_item = SETUP_ITEMS-1;
      setup_cursor = MAX_SETUP_ITEMS-1;
      setup_offset = SETUP_ITEMS - MAX_SETUP_ITEMS;
      refresh_setup_window();
   }
   return KEY_NOTHING;
}

int setup_down()
{
   if (selected_setup_item < SETUP_ITEMS-1)
   {
      do
      {
         selected_setup_item++;
         if (setup_cursor < MAX_SETUP_ITEMS-1) setup_cursor++;
         else
         if (selected_setup_item >= MAX_SETUP_ITEMS) { setup_offset = selected_setup_item - MAX_SETUP_ITEMS + 1; continue; }
      } while (is_grayed_out(selected_setup_item + SI_START));
      if (selected_setup_item > SETUP_ITEMS-1)
	 { selected_setup_item = prev_selected_setup_item; setup_cursor = prev_setup_cursor; setup_offset = prev_setup_offset; }
      refresh_setup_window();
      prev_selected_setup_item = selected_setup_item; prev_setup_cursor = setup_cursor; prev_setup_offset = setup_offset;
   }
   else
   {
      selected_setup_item = 0;
      setup_cursor = 0;
      setup_offset = 0;
      refresh_setup_window();
   }
   return KEY_NOTHING;
}

int run_script(char *script_fname)
{
   char cmd[MAX_PATH];
   char cdcmd[MAX_PATH];

   sprintf(cmd, "%s%s/%s", config.path, SCRIPTS_DIR, script_fname);

   FILE *cf;

   if (access(cmd, F_OK) != 0) return SCRIPT_NOT_FOUND;

   cf = fopen(CONTROL_SCRIPT_FNAME, "wt");
   if (cf)
   {
      if (fputs("opwd=`pwd`\n", cf) < 0) { fclose(cf); remove(CONTROL_SCRIPT_FNAME); return SCRIPT_FAILED; }
      sprintf(cdcmd, "cd %s%s\n", config.path, SCRIPTS_DIR);
      if (fputs(cdcmd, cf) < 0) { fclose(cf); remove(CONTROL_SCRIPT_FNAME); return SCRIPT_FAILED; }
      if (fputs(cmd, cf) < 0) { fclose(cf); remove(CONTROL_SCRIPT_FNAME); return SCRIPT_FAILED; }
      if (fputs("\ncd $opwd\n", cf) < 0) { fclose(cf); remove(CONTROL_SCRIPT_FNAME); return SCRIPT_FAILED; }
      sprintf(cmd, "\ntouch %s\n", CONTROL_TAG_FNAME);
      if (fputs(cmd, cf) < 0) { fclose(cf); remove(CONTROL_SCRIPT_FNAME); return SCRIPT_FAILED; }
      fclose(cf);

      if (access(CONTROL_TAG_FNAME, F_OK) == 0) remove(CONTROL_TAG_FNAME);

      if (!telnetd_working)
      {
	 telnetd_working = is_telnetd();
	 if (!telnetd_working)
	    if (!run_telnetd()) return SCRIPT_FAILED;
	 telnetd_working = 1;
      }

      sprintf(cmd, "chmod a+x %s;%s", CONTROL_SCRIPT_FNAME, CONTROL_SCRIPT_FNAME);
      if (system(cmd) == -1) { remove(CONTROL_SCRIPT_FNAME); return SCRIPT_FAILED; }

      int wait = SCRIPT_MAX_TIME;
      while (wait > 0)
      {
	 if (access(CONTROL_TAG_FNAME, F_OK) == 0) wait = 0;
	 SDL_Delay(SCRIPT_TIME_STEP); wait-=SCRIPT_TIME_STEP;
      }

      if (access(CONTROL_SCRIPT_FNAME, F_OK) == 0) remove(CONTROL_SCRIPT_FNAME);
      if (access(CONTROL_TAG_FNAME,    F_OK) == 0) remove(CONTROL_TAG_FNAME);
   }
   else return SCRIPT_FAILED;

   return SCRIPT_SUCCESS;
}

int run_pre_script(int id)
{
   int i, ln;
   char sname[128], cmd[MAX_PATH];

   if ((!config.pre_scripts) || (config.target1.id == 0)  || divide_now) return SCRIPT_SUCCESS;

   sprintf(sname, "%s.sh", targets[id].product[0]?targets[id].product:targets[id].name);

   ln = strlen(sname);
   for (i=0; i < ln; i++)
       if (((sname[i]>='A') && (sname[i]<='Z')) || ((sname[i]>='a') && (sname[i]<='z')) ||
           ((sname[i]>='0') && (sname[i]<='9')) || (sname[i]=='-') || (sname[i]=='.')) ;
       else sname[i] = '_';

   sprintf(cmd, "%s%s/%s", config.path, SCRIPTS_DIR, sname);
   if (access(cmd, F_OK) != 0) return SCRIPT_NOT_FOUND;

   show_SCRIPT();

   int ret = run_script(sname);
   close_video_mode();
   return ret;
}

void set_io_buffer_size(unsigned val)
{
   static unsigned *pvr_recinstance;

   if (val > 0)
   {
      spIPvr_GetRecInstance(0, &pvr_recinstance);
      if (pvr_recinstance) *(pvr_recinstance + 0x2C/4) = val;
   }
}

int is_target_accessible(int nr)
{
   struct statfs fs;

   if ((nr < 0) || (nr >= max_targets)) return 0;

   if ((config.createintdirs) && (targets[nr].usesintmem == 1))
      mkdir(targets[nr].path, ACCESSPERMS);

   if (access(targets[nr].path, F_OK) != 0) return 0;

   if (targets[nr].usesintmem == 1)
      if (statfs(targets[nr].path, &fs) == 0)
	 if (fs.f_type == FS_RFS)
	    if (!config.useintmem) return 0;

   return 1;
}

static int create_recname()
{
   FILE *testwrite;
   char s[MAX_PATH];
   int nr;
   int the_best = -1;
   long max_free = 0L;

   config.current_target = 0;
   network_drive = 0;
   internal_memory = 0;
   epg_probes = 0L;
   next_probe = 0L;

   current_divide = config.divide;
   if (current_divide == DIVIDING_AUTO) current_divide = DIVIDING_OFF;

   refresh_config_targets();

   if (no_dir1 == 0)
   {
      if (config.target1.id == 0)
      {
         for (nr=1; nr < max_targets; nr++)
         {
	     if (is_target_accessible(nr))
	     {
		if (statfs(targets[nr].path, &fs) == 0)
                {
		   if (fs.f_type != FS_NTFS)
		   {
		      if (fs.f_bavail > max_free)
		      {
			 max_free = fs.f_bavail;
			 if (max_free >= free_blocks_minimum) the_best = nr;
		      }
		   }
                }
	     }
         }
         if (the_best == -1) { no_dir1 = 0; return -1; }
	 config.current_target = the_best;
      }
      else
      {
	 if (is_target_accessible(config.target1.id))
	 {
	    if (is_NTFS(targets[config.target1.id].path)) { no_dir1 = 1; return -3; }
	    config.current_target = config.target1.id;
	 }
	 else { no_dir1 = 1; return -1; }
      }
   }
   else if (no_dir1 == 1)
   {
      if (config.target2.id == 0)
      {
         for (nr=1; nr < max_targets; nr++)
         {
	     if (is_target_accessible(nr))
	     {
		if (statfs(targets[nr].path, &fs) == 0)
                {
		   if (fs.f_type != FS_NTFS)
		   {
		      if (fs.f_bavail > max_free)
		      {
			 max_free = fs.f_bavail;
			 if (max_free >= free_blocks_minimum) the_best = nr;
		      }
		   }
                }
	     }
         }
         if (the_best == -1) { no_dir1 = 0; return -2; }
	 config.current_target = the_best;
      }
      else
      {
	 if (is_target_accessible(config.target2.id))
	 {
	    if (is_NTFS(targets[config.target2.id].path)) { no_dir1 = 0; return -3; }
	    config.current_target = config.target2.id;
	 }
	 else { no_dir1 = 0; return -2; }
      }
   }

   if ((fs.f_type == FS_CIFS) || (fs.f_type == FS_SMB) || (fs.f_type == FS_NFS)) network_drive = 1;
   if ((fs.f_type == FS_RFS) && (targets[config.current_target].usesintmem)) internal_memory = 1;

   if (config.divide == DIVIDING_AUTO)
   {
      if (network_drive) current_divide = DIVIDING_OFF;
      else if (fs.f_type == FS_RFS) current_divide = DIVIDING_1GB;
      else if (fs.f_type == FS_FAT) current_divide = DIVIDING_2GB;
      else current_divide = DIVIDING_4GB;
   }

   nr = MAX_DVR_SUBDIRS;
   config.DVRnr = 0;
   while (nr > 0)
   {
      sprintf(s, "%s/%s%d", targets[config.current_target].path, DVR_SUBDIR_NAME, nr);
      if (access(s, F_OK) == 0)
	 { config.DVRnr = nr; nr = 0; }
      else nr--;
   }

   if (config.DVRnr > 0)
      sprintf(s, "%s/%s%d", targets[config.current_target].path, DVR_SUBDIR_NAME, config.DVRnr);
   else strcpy(s, targets[config.current_target].path);

   EPG_get_data();

   for (nr=1; nr < 999; nr++)
   {
	sprintf(ts_filename,  ts_file_pattern, s, recpartname, nr);
	if (access(ts_filename, F_OK) != 0)
	{
	   if (config.testwrite)
	   {
              testwrite = fopen(ts_filename, "wb");
	      if (testwrite)
	      {
	         fclose(testwrite);
	         remove(ts_filename);
	      }
	      else continue;
	   }
	   sprintf(mdb_filename, mdb_file_pattern, s, recpartname, nr);
	   return nr;
	}
   }
   return (config.current_target == config.target2.id)?-2:-1;
}

int delete_empty(char *ts_fname, char *mdb_fname)
{
   FILE *f;

   int del = 0;
   if (ts_fname[0])
   {
      if (config.del_empty_ts)
      {
	 f = fopen(ts_fname, "rb");
	 if (f)
	 {
	    fseek(f, 0, SEEK_END);
	    if (ftell(f) == 0)
	    {
	       fclose(f);
	       remove(ts_fname);
	       if ((!config.del_mdb) && (mdb_fname[0])) remove(mdb_fname);
	       del = 1;
	    }
	    else fclose(f);
	 }
      }
   }
   return del;
}

int pause_video()
{
   unsigned int tDpHndl;
   int ret;

   ret = spIDp_Open(DP_INST0, &tDpHndl);
   if ((ret != 0) || (tDpHndl == 0)) return KEY_PAUSE;
   pause_state = (pause_state == PAUSE_OFF) ? PAUSE_ON : PAUSE_OFF;
   spIDp_SetFreeze(tDpHndl, pause_state);

   spIDp_Close(tDpHndl);

   return KEY_NOTHING;
}

void *mesuring_thread_proc(void *ptr)
{
   measuring_thread_working = 1;
   while (button_is_pressed && sustain_measuring_thread)
   {
      sleep(1);
      pressed_time++;
      if (pressed_time == 3)
      {
	 if (measuring_key == KEY_MUTE)
	 {
	    if (!panel_off) esaving_panel_off();
	 }
	 sustain_measuring_thread = 0;
      }
   }
   measuring_thread_working = 0;
   pthread_exit((void*)ptr);
}

void start_measuring_thread(int key)
{
   measuring_key = key;
   if (measuring_thread_working) return;

   pressed_time = 0;

   sustain_measuring_thread = 1;
   measuring_thread_ID = pthread_create(&measuring_pthr, NULL, &mesuring_thread_proc, NULL);
   if (measuring_thread_ID == 0) measuring_thread_working = 0;
}

void stop_measuring_thread()
{
   measuring_key = KEY_NOTHING;
   if (measuring_thread_working)
   {
      sustain_measuring_thread = 0;
      int timeout=8;
      while ((measuring_thread_working) && timeout--) sleep(1);
      if (timeout <= 0) pthread_kill(measuring_pthr, SIGALRM);
   }
}

static int button_pressed(int key)
{
  int deleted;
  int i;
  char s[MAX_PATH];

  int hardkey = 1;
  if (key < 0) { hardkey = 0; key = -key; }
  else
  {
     button_is_pressed = 1;
  }

  int key_really_pressed = key;

  if (emergency_stop == 2) { emergency_stop = 0; close_video_mode(); }
  if (esaving_restore > 0)
  {
     esaving_set_option(TCTV_OPTION_ESAVING, cur_saving_mode);
     esaving_restore--;
  }

  switch (key)
  {
	case KEY_POWER:
	case KEY_BEZEL_POWER:
	if (question)
	{
	   OSD_cls();
	   question = 0;
	   close_video_mode();
	   key = KEY_NOTHING;
	   break;
	}
	if (setup_menu)	{ close_setup_window(0); break; }
	if ((is_recording) && (autostop != 0))
	{
	   if (config.poweroff) config.poweroff = 0; else config.poweroff = 1;
	   key = show_WARNING(config.poweroff?MSG_POFF_ON:MSG_POFF_OFF, 0);
	   clear_WARNING();
	   break;
	}

	case KEY_STOP:
	if (question || setup_menu) { key = KEY_NOTHING; break; }
	if (is_recording)
	{
	   if (config.teletext)
	      if (mode_teletext > 0) { key = KEY_NOTHING; break; }

		if (!divide_now)
		{
		   if (video_muted)	esaving_togglevideo();
		   if (panel_off)	esaving_panel_on();
		}

		is_recording = 0;

		if (!divide_now)
		{
		   i = 40;
		   while (monitoring_now && --i) SDL_Delay(100);
		   SDL_Delay(20);
		}

		if ((key == KEY_STOP) && (!divide_now))
		{
		    if (recdispmode != RECDISPMODE_HIDDEN)
		    {
		       set_icon(ICON_STOPREC);
		    }
		}

		spI_Stream_Record_Stop();

		RECfile_MBsize = get_file_MBsize(ts_filename);
		prev_files_size += RECfile_MBsize;

		if (!divide_now)
		{
		   if (recdispmode == RECDISPMODE_INFOPANE) infopane_hide();
		   recdispmode = RECDISPMODE_NORMAL;
		}

		if ((mdb_filename[0]) && (config.del_mdb))
		{
		   remove(mdb_filename);
		   mdb_filename[0]=0;
		}

		deleted = delete_empty(ts_filename, mdb_filename);

		if ((epg_probes > 0) && (recpartname_best[0]))
		{
		   if (config.DVRnr > 0)
		      sprintf(s, "%s/%s%d", targets[config.current_target].path, DVR_SUBDIR_NAME, config.DVRnr);
		   else strcpy(s, targets[config.current_target].path);

		   sprintf(new_filename,  ts_file_pattern, s, recpartname_best, part_number+1);

		   if (access(ts_filename, F_OK) == 0)
		      if (access(new_filename, F_OK) != 0)
			 if (rename(ts_filename, new_filename) == 0)
			 {
			    strcpy(ts_filename, new_filename);
			 }
		}

		if (!deleted)
		{
	           i = rec_history;
		   if (i >= MAX_REC_HISTORY) i--;
		   strcpy(rec_hist_files[i].ts_filename, ts_filename);
		   strcpy(rec_hist_files[i].mdb_filename, mdb_filename);
		   rec_history = i+1;
		}

		if (divide_now) return KEY_NOTHING;

		if (key == KEY_STOP)
		{
		    key = stop_REC(deleted);
		    if (config.summary && (!deleted)) show_summary();
		}
		no_dir1 = 0;
	}
	break;

	case KEY_RECORD:
	if (question) { key = KEY_NOTHING; break; }
	if ((config.teletext) && (!divide_now))
	   if (mode_teletext > 0) { key = KEY_NOTHING; break; }
	if ((!divide_now) && (is_recording))
	{
	   switch(recdispmode)
	   {
	      case RECDISPMODE_NORMAL:
		if (config.OSD == 0) break;
		recdispmode = RECDISPMODE_INFOPANE;
		infopane_show(config.OSD_effects);
		break;
	      case RECDISPMODE_INFOPANE:
		recdispmode = RECDISPMODE_NORMAL;
		infopane_hide();
		break;
	      default:
		show_REC();
		recdispmode = RECDISPMODE_NORMAL;
		break;
	   }
	   key = KEY_NOTHING;
	}
	else
	if ((!divide_now) && (prev_key == KEY_MENU)) { if (!setup_menu) key = setup(); }
	else
	if ((!divide_now) && (config.record_key_lock)) key = show_WARNING(MSG_LOCKED, 1);
	else if (!setup_menu)
	{
	   init_targets();
	   refresh_config_targets();

	   if ((!divide_now) && (config.pre_scripts) && (config.target1.id > 0))
	   {
	      if (run_pre_script(config.target1.id) == SCRIPT_FAILED)
	      {
		 key = show_WARNING(MSG_SCRIPT_FAILED, 1);
		 break;
	      }
	   }

           int cn = create_recname();
	   if (cn > 0)
	   {
	      if (statfs(targets[config.current_target].path, &fs) != 0) fs.f_type = 0;

	      RECfile_MBsize = 0L;
	      divide_warned = 0;
	      lowspace_warned = 0;
	      lowspace_now = 0;
	      emergency_stop = 0;

	      rec_partstarttime = time(NULL);
	      is_recording = 1;
	      if (!divide_now)
	      {
		 rec_history = 0;
		 part_number = 0;
		 prev_files_size = 0L;
		 rec_starttime = time(NULL);
	         recdispmode = RECDISPMODE_NORMAL;
		 strcpy(freespace, "?");
		 totalfree_step = 0;
		 epg_step = EPG_PERIOD - EPG_FIRST_PERIOD;
		 infopane_update = 0;
		 bitrate = 0;
		 free_minutes = 0;
		 free_level = -1;
		 monitoring_now = 0;
		 autostop = 0;
		 recwarnings_step = 0;
		 recwarned = 0;
		 EPG_progname_changed = 0;
		 EPG_namebuffer[0] = 0;
	      }
	      else part_number++;

	      spI_Stream_Record(ts_filename);

	      if (!divide_now)
	      {
	         init_video_mode();
	         key = show_REC();
	      }
	      else return KEY_NOTHING;

	      divide_now = 0;

	      rec_timer_id = SDL_AddTimer(REC_TIMER_RATE, rec_monitoring, 0);

	      if ((config.infopane == INFOPANE_AUTO) && (config.OSD) && (recdispmode == RECDISPMODE_NORMAL))
	      {
		 recdispmode = RECDISPMODE_INFOPANE;
		 infopane_show(config.OSD_effects);
	      }
	   }
	   else
	   {
		key = show_WARNING((cn==-1)?MSG_NOREC1:((cn==-2)?MSG_NOREC2:MSG_NTFS), 1);
	   }
	}
	break;

	case KEY_COL_RED:
	if (question || setup_menu) { key = KEY_NOTHING; break; }
	if ((!is_recording) && (prev_key == KEY_STOP) && (rec_history > 0))
	   key = show_QUESTION(MSG_DELLASTREC);
	break;

	case KEY_FORWARD:
	if (question || setup_menu) { key = KEY_NOTHING; break; }
	if (is_recording)
	{
	   if (autostop < MAX_AUTOSTOP)
	   {
	      if (autostop == -1) autostop = 0;
	      else if (autostop < AUTOSTOP_STEP2) autostop += AUTOSTOP_STEP1;
	      else autostop += AUTOSTOP_STEP2;
	   }
	   if (autostop > MAX_AUTOSTOP) autostop = MAX_AUTOSTOP;
	   if (autostop > 0) autostop_time = time(NULL) + autostop*60;
	   show_AUTOSTOP();
	   key = KEY_NOTHING;
	}
	break;

	case KEY_REWIND:
	if (question || setup_menu) { key = KEY_NOTHING; break; }
	if (is_recording)
	{
	   if (autostop == 0) autostop = -1;
	   else
	   if (autostop > 0)
	   {
	      if (autostop > AUTOSTOP_STEP2) autostop -= AUTOSTOP_STEP2; else autostop -= AUTOSTOP_STEP1;
	   }
	   if (autostop > 0) autostop_time = time(NULL) + autostop*60;
	   show_AUTOSTOP();
	   key = KEY_NOTHING;
	}
	break;

	case KEY_PAUSE:
	if (config.freeze == FREEZE_AUTO)
	{
	   if (mode_mplay) break;
	   if (mode_internet) break;
	}
	if (config.freeze != FREEZE_OFF) key = pause_video();
	if (question || setup_menu) { key = KEY_NOTHING; break; }
	break;

	case KEY_PLAY:
	if (question || setup_menu) { key = KEY_NOTHING; break; }
	if (is_recording) { show_AUTOSTOP(); key = KEY_NOTHING; }
	else if (pause_state == PAUSE_ON)
	{
	   if (mode_mplay) break;
	   if (mode_internet) break;
	   key = pause_video();
	}
	break;

	case KEY_CUR_UP:
	if (question) { key = KEY_NOTHING; break; }
	if (setup_menu) key = setup_up();
	break;

	case KEY_CUR_DOWN:
	if (question) { key = KEY_NOTHING; break; }
	if (setup_menu) key = setup_down();
	break;

	case KEY_CUR_LEFT:
	if (question)
	{
	   question_answer = 1;
	   question_update();
	   key = KEY_NOTHING;
	}
	if (setup_menu) key = setup_left();
	break;

	case KEY_CUR_RIGHT:
	if (question)
	{
	   question_answer = 0;
	   question_update();
	   key = KEY_NOTHING;
	}
	if (setup_menu) key = setup_right();
	break;

	case KEY_ENTER:
	if (question)
	{
	   OSD_cls();

	   if ((question_answer == 1) && (rec_history > 0))
	   {
	      for (i=0; i < rec_history; i++)
	      {
		  remove(rec_hist_files[i].ts_filename);
		  if (!config.del_mdb) remove(rec_hist_files[i].mdb_filename);
	      }
	      show_WARNING(MSG_REC_DELETED, 1);
	   }

	   question = 0;
	   close_video_mode();
	   key = KEY_NOTHING;
	   break;
	}
	if (setup_menu)
	{
	   close_setup_window(1);
	   key = KEY_NOTHING;
	}
	break;

	case KEY_RETURN:
	if (question)
	{
	   OSD_cls();
	   question = 0;
	   close_video_mode();
	   key = KEY_NOTHING;
	   break;
	}
	if (setup_menu)
	{
	   close_setup_window(0);
	   key = KEY_MENU;
	   key_really_pressed = KEY_MENU;
	}
	break;

	case KEY_EXIT:
	mode_mplay = 0;
	mode_internet = 0;
	if (question)
	{
	   OSD_cls();
	   question = 0;
	   close_video_mode();
	   key = KEY_NOTHING;
	   break;
	}
        if (setup_menu)
	{
	   close_setup_window(0);
	}
        else
	if ((config.teletext) && (mode_teletext > 0)) break;
	else
	if (is_recording) switch(recdispmode)
	{
	   case RECDISPMODE_NORMAL:
		OSD_cls();
		recdispmode = RECDISPMODE_HIDDEN;
		break;
	   case RECDISPMODE_INFOPANE:
		infopane_hide();
		OSD_cls();
		recdispmode = RECDISPMODE_HIDDEN;
		break;
	   default:
		break;
	}
	break;

	case KEY_MENU:
	if (setup_menu) close_setup_window(0);
	if (question) key = KEY_NOTHING;
	break;

	case KEY_SOURCE:
	case KEY_BEZEL_SOURCE:
	if (question || setup_menu) { key = KEY_NOTHING; break; }
	if (is_recording) key = show_NO();
	else
	if ((prev_key != KEY_SOURCE) && (prev_key != KEY_BEZEL_SOURCE))
	{
	   if (config.source == SOURCE_SCRIPT)
	   {
	      int script_ret = run_script(SOURCE_SCRIPT_FNAME);
	      if (script_ret == SCRIPT_NOT_FOUND) show_WARNING(MSG_NOSCRIPT, 1);
	      else if (script_ret == SCRIPT_FAILED) show_WARNING(MSG_SCRIPT_FAILED, 1);
	      key = KEY_NOTHING;
	   }
	   else
	   key = sources[config.source].value;
	}
	break;

	case KEY_FAV_CH:
	if (question || setup_menu) { key = KEY_NOTHING; break; }
	if (prev_key != KEY_STOP)
	{
	   if (config.favch == FAVCH_SCRIPT)
	   {
	      int script_ret = run_script(FAVCH_SCRIPT_FNAME);
	      if (script_ret == SCRIPT_NOT_FOUND) show_WARNING(MSG_NOSCRIPT, 1);
	      else if (script_ret == SCRIPT_FAILED) show_WARNING(MSG_SCRIPT_FAILED, 1);
	      key = KEY_NOTHING;
	   }
	   else
	   {
	      key = favchs[config.favch].value;
	      if (key == KEY_FAV_CH)
	         if (is_recording) key = show_NO();
	   }
	}
	else if (is_recording) key = show_NO();
	break;

	case KEY_MUTE:
	if (question || setup_menu) { key = KEY_NOTHING; break; }
	if (config.mute == 1) start_measuring_thread(KEY_MUTE);
	break;

	case KEY_TTX_MIX:
	if (question || setup_menu) { key = KEY_NOTHING; break; }
	if (is_recording)
	{
	   if (!config.teletext)
	   {
	      key = show_NO();
	      if (recdispmode == RECDISPMODE_INFOPANE) infopane_show(0);
	      break;
	   }
	   if (recdispmode == RECDISPMODE_INFOPANE) infopane_hide();
	   OSD_cls();
	   recdispmode = RECDISPMODE_HIDDEN;
	}
	mode_teletext++; if (mode_teletext > 3) mode_teletext = 1;
	break;

	case KEY_TV:
	if (question) { key = KEY_NOTHING; break; }
	if (setup_menu) { close_setup_window(0); break; }
	if (mode_teletext > 0) mode_teletext = 0;
	else
	if (is_recording)
	{
	   key = show_NO();
	   if (recdispmode == RECDISPMODE_INFOPANE) infopane_show(0);
	}
	break;

	case KEY_PRE_CH:
	case KEY_BEZEL_P_DOWN:
	case KEY_BEZEL_P_UP:
	case KEY_P_UP:
	case KEY_P_DOWN:
	case KEY_NUM0:
	case KEY_NUM1: case KEY_NUM2: case KEY_NUM3:
	case KEY_NUM4: case KEY_NUM5: case KEY_NUM6:
	case KEY_NUM7: case KEY_NUM8: case KEY_NUM9:
	if (question || setup_menu) { key = KEY_NOTHING; break; }
	if (config.teletext) if (mode_teletext > 0) break;

	case KEY_MEDIA_P:
	if ((key == KEY_MEDIA_P) && (!question))
	{
	   rec_history = 0;
	   mode_mplay = 1;
	}
	case KEY_CH_LIST:
	case KEY_CONTENT:	
	case KEY_HDMI1: case KEY_HDMI2: case KEY_HDMI3: case KEY_HDMI4:
	case KEY_VGA: case KEY_EXT1: case KEY_EXT2: case KEY_COMPONENT:
	case KEY_HDMI_TOGGLE: case KEY_SERVICE: case KEY_DIAG1:
	if (question || setup_menu) { key = KEY_NOTHING; break; }
	if (is_recording)
	{
	   key = show_NO();
	   if (recdispmode == RECDISPMODE_INFOPANE) infopane_show(0);
	}
	break;

	case KEY_INTERNET:
	     if (question || setup_menu) { key = KEY_NOTHING; break; }
	     mode_internet = 1;
	case KEY_GUIDE:
	case KEY_INFO:
	case KEY_SUBT:
	if (setup_menu) key = KEY_NOTHING;
	break;

	default:
	if (question) { if ((key != KEY_VOL_PLUS) && (key != KEY_VOL_MINUS)) key = KEY_NOTHING; break; }
	break;
  }

  prev_key = key_really_pressed;

  if (key != KEY_NOTHING)
     if (hardkey)
	if (panel_off)
	{
	   panel_off = 0;
	   esaving_restore = 1;
	}
  return(key);
}

static void __attribute__ ((unused)) button_released(void)
{
   stop_measuring_thread();
   button_is_pressed = 0;
} 

void injection_press(int a, int key, int arg_r2) __attribute__ ((noinline));
void injection_press(int a, int key, int arg_r2)
{
	asm volatile(
		"STMFD   SP!, {R0,R2,R3,R12}\n"
	);

	button_pressed(key);

	asm volatile(
		"MOV     R1, R0\n"
		"MOV     R6, R0\n"
		"LDMFD   SP!, {R0,R2,R3,R12}\n"
		"LDMFD   SP!, {R4,LR}\n"
		"MOV     R7, R2\n"
		"LDR     R3, [R8]\n"
		"LDR     PC, =(_ZN9KeyCommon17SendKeyPressInputEii + 0x18 )\n"
	);
}

void injection_release(int a, int key, int arg_r2) __attribute__ ((noinline));
void injection_release(int a, int key, int arg_r2)
{
	asm volatile(
		"STMFD   SP!, {R0-R12, R14}\n"
	);

	button_released();
	
	asm volatile(
		"LDMFD   SP!, {R0-R12, R14}\n"
		"LDMFD   SP!, {R4,LR}\n"
		"MOV     R7, R1\n"
		"LDR     R3, [R5]\n"
		"LDR     PC, =(_ZN9KeyCommon19SendKeyReleaseInputEi + 0x18 )\n"
	);
}

void init_config(cfgstruct *loader_config)
{
    max_targets = 0;
    memcpy(&config, loader_config, sizeof(cfgstruct));
    telnetd_working = is_telnetd();
    init_targets();
    setup_menu  = 0;
    mode_teletext = 0;
    mode_mplay	= 0;
    mode_internet = 0;
    no_dir1     = 0;
    is_recording= 0;
    recdispmode = RECDISPMODE_NORMAL;
    prev_key    = -1;
    question    = 0;
    rec_history = 0;

    init_OSD();
    esaving_init();
    set_io_buffer_size(config.io_buffer);
}

